#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin

from ads.common import utils
from ads.dataset import logger
from ads.dataset.helper import down_sample, up_sample, get_fill_val
from ads.dataset.progress import DummyProgressBar
from ads.dataset.recommendation import Recommendation
from ads.type_discovery.typed_feature import (
    ContinuousTypedFeature,
    DiscreteTypedFeature,
    OrdinalTypedFeature,
    CategoricalTypedFeature,
)


class RecommendationTransformer(TransformerMixin):
    def __init__(
        self,
        feature_metadata=None,
        correlation=None,
        target=None,
        is_balanced=False,
        target_type=None,
        feature_ranking=None,
        len=0,
        fix_imbalance=True,
        auto_transform=True,
        correlation_threshold=0.7,
    ):
        self.feature_metadata_ = feature_metadata
        self.correlation_ = correlation
        self.target_ = target
        self.target_type_ = target_type
        self.feature_ranking_ = feature_ranking
        self.fill_nan_dict_ = {}
        self.fill_na_target_ = None
        self.drop_columns = []
        self.reco_dict_ = None
        # self.combine_columns = []
        self.actions_performed_ = []
        self.is_balanced = is_balanced
        self.ds_len = len
        self.fix_imbalance = fix_imbalance
        self.balancing_strategy = None
        self.auto_transform = auto_transform
        self.correlation_threshold = correlation_threshold

    def __repr__(self):
        if len(self.actions_performed_) == 0:
            return "No recommendations suggested"
        return "\n".join(self.actions_performed_)

    @staticmethod
    def _build_recommendation(
        recommendations,
        recommendation_type,
        column_names,
        message,
        actions,
        recommended_action=None,
    ):
        if len(column_names) != 0:
            if recommendation_type in ["constant_column"]:
                recommendations[recommendation_type] = column_names
            else:
                if recommendation_type not in recommendations:
                    recommendations[recommendation_type] = {}
                for column in column_names:
                    if column not in recommendations:
                        recommendations[recommendation_type][column] = {}
                    recommendations[recommendation_type][column]["Message"] = message
                    recommendations[recommendation_type][column]["Action"] = actions
                    recommendations[recommendation_type][column]["Selected Action"] = (
                        recommended_action
                        if recommended_action is not None
                        else actions[0]
                    )

    def _get_recommendations(self, df):
        recommendations = {}
        # constant columns
        constant_columns = df.columns.values[df.apply(pd.Series.nunique) == 1]
        if self.target_ in constant_columns:
            raise ValueError(
                "Unable to continue with transformation. Target column is constant. Build the dataset "
                "by choosing a different target."
            )
        self._build_recommendation(
            recommendations,
            "constant_column",
            constant_columns,
            "Constant Column",
            ["Drop"],
        )

        # primary key
        for column in df.columns.values[
            df.apply(lambda x: x.nunique() / len(x) > 0.99)
        ]:
            # exclude columns of dtype object from primary key check as they could be columns like zipcode, credit card,
            # etc,. which are mostly unique but carries useful information
            if (
                (
                    "constant_column" not in recommendations
                    or column not in recommendations["constant_column"]
                )
                and column != self.target_
                and df[column].dtype.name.startswith("int")
            ):
                self._build_recommendation(
                    recommendations,
                    "primary_key",
                    [column],
                    "Contains mostly unique values({0:.2%})".format(
                        df[column].nunique() / len(df[column])
                    ),
                    ["Drop", "Do nothing"],
                    "Drop",
                )

        self.feature_metadata_[self.target_] = self.target_type_

        for column in df.columns.values[df.isnull().any()]:

            # filter out columns that were discovered as constant or primary key columns in the previous step,
            # as they would get dropped before imputation
            if (
                "constant_column" not in recommendations
                or column not in recommendations["constant_column"]
            ) and (
                "primary_key" not in recommendations
                or column not in recommendations["primary_key"]
            ):
                null_counts = df[column].isnull().sum()
                null_ratio = null_counts / len(df[column])
                self._get_na_action(recommendations, column, null_counts, null_ratio)

        if self.correlation_ is not None:
            #  highly correlated features
            corr_features = []
            if not isinstance(self.correlation_, list):
                self.correlation_ = [self.correlation_]
            for corr in self.correlation_:
                high_corr_var = np.where(corr > self.correlation_threshold)
                corr_features.extend(
                    [
                        (corr.index[x], corr.columns[y], corr.iat[x, y])
                        for x, y in zip(*high_corr_var)
                        if x != y and x < y
                    ]
                )
            if len(corr_features) > 0:
                # Apply all recommendations so far
                self.reco_dict_ = recommendations
                df = self._transform(df)

                for corr_feature in corr_features:
                    if (
                        self.target_ not in corr_feature
                        and corr_feature[0] in df.columns.values
                        and corr_feature[1] in df.columns.values
                        and corr_feature[0] != corr_feature[1]
                    ):
                        feature1_rank_and_score = self.feature_ranking_[
                            self.feature_ranking_["features"] == corr_feature[0]
                        ]["scores"]
                        rank1, score1 = (
                            feature1_rank_and_score.index[0],
                            feature1_rank_and_score.values[0],
                        )
                        feature2_rank_and_score = self.feature_ranking_[
                            self.feature_ranking_["features"] == corr_feature[1]
                        ]["scores"]
                        rank2, score2 = (
                            feature2_rank_and_score.index[0],
                            feature2_rank_and_score.values[0],
                        )
                        # if any of the features is top ranked or ranked similar, combine
                        # if ((len(df.columns) - rank1 + 1) / len(df.columns) > 0.7 or \
                        #         (len(df.columns) - rank2 + 1) / len(df.columns) > 0.7):
                        #     selected_action = "Combine with " + corr_feature[1]

                        # suggest dropping the column with lesser importance
                        if score1 > score2:
                            selected_action = "Drop " + corr_feature[1]
                        else:
                            selected_action = "Drop " + corr_feature[0]
                        # check if the corr_feature isn't target
                        self._build_recommendation(
                            recommendations,
                            "strong_correlation",
                            [corr_feature[0]],
                            "Strongly correlated with "
                            + corr_feature[1]
                            + "({0:.2%}.".format(corr_feature[2])
                            + ")",
                            [
                                "Drop " + corr_feature[0],
                                "Drop " + corr_feature[1],
                                # "Combine with " + corr_feature[1],
                                "Do nothing",
                            ],
                            selected_action,
                        )

        if isinstance(self.target_type_, DiscreteTypedFeature):
            unique_vals = list(self.target_type_.meta_data["internal"]["counts"].keys())
            # binary classification dataset, suggest setting a positive class only if the values are not True/False
            if len(unique_vals) == 2 and True not in unique_vals:
                unique_vals.append("Do nothing")

                # for auto transform, do not suggest a default unless it is one of the known positive classes
                selected_action = unique_vals[2]

                # find positive label if recommendations can be manually updated
                if not self.auto_transform:
                    pos_vals = ["Y", "YES", "y", "Yes", "yes", "1", "true"]
                    pos_label_index = np.where(np.isin(pos_vals, unique_vals) == True)[
                        0
                    ]
                    selected_action = (
                        pos_vals[pos_label_index[0]]
                        if len(pos_label_index) > 0
                        else unique_vals[0]
                    )

                self._build_recommendation(
                    recommendations,
                    "positive_class",
                    [self.target_],
                    "Set Positive Class",
                    unique_vals,
                    selected_action,
                )
            # check if dataset is imbalanced
            if not self.is_balanced and self.fix_imbalance:
                target_value_counts = df[self.target_].value_counts()
                minority_class_len = min(
                    target_value_counts.iteritems(), key=lambda k: k[1]
                )[1]
                majority_class_len = max(
                    target_value_counts.iteritems(), key=lambda k: k[1]
                )[1]
                minor_majority_ratio = minority_class_len / majority_class_len

                # Suggest down sampling if minor_majority_ratio is 1:1000
                suggested_sampling = (
                    "Do nothing"
                    if len(df) <= utils.MAX_LEN_FOR_UP_SAMPLING
                    else "Down-sample"
                    if minor_majority_ratio >= utils.MIN_RATIO_FOR_DOWN_SAMPLING
                    else "Do nothing"
                )
                self._build_recommendation(
                    recommendations,
                    "fix_imbalance",
                    [self.target_],
                    "Imbalanced Target({0:.2%})".format(minor_majority_ratio),
                    ["Do nothing", "Down-sample", "Up-sample"],
                    suggested_sampling,
                )

        return recommendations

    def fit(self, X):
        self.reco_dict_ = self._get_recommendations(X)
        return self

    def fit_transform(self, X, y=None, **fit_params):
        if self.reco_dict_ is None:
            self.fit(X)
        return self.transform(X, fit_transform=True, update_transformer_log=True)

    def transform(
        self,
        X,
        progress=DummyProgressBar(),
        fit_transform=False,
        update_transformer_log=False,
    ):
        df = self._transform(
            X,
            progress=progress,
            fit_transform=fit_transform,
            update_transformer_log=update_transformer_log,
        )

        # cleanup unused objects
        if hasattr(self, "reco_dict_") and not fit_transform:
            del self.reco_dict_
            del self.feature_metadata_
            del self.correlation_
            del self.feature_ranking_
        return df

    def transformer_log(self, action):
        """local wrapper to both log and record in the actions_performed array"""
        logger.info(action)
        self.actions_performed_.append(action)

    def _transform(
        self,
        X,
        progress=DummyProgressBar(),
        fit_transform=False,
        update_transformer_log=False,
    ):
        if hasattr(self, "reco_dict_") and len(self.reco_dict_) > 0:
            self.drop_columns = []
            # self.combine_columns = []
            # self.fill_nan_dict_ = {}
            columns_to_drop = []
            for recommendation_type_index in range(
                0, len(Recommendation.recommendation_types)
            ):
                recommendation_type = Recommendation.recommendation_types[
                    recommendation_type_index
                ]
                if recommendation_type in self.reco_dict_:
                    if recommendation_type_index == 0:
                        columns_to_drop = self.reco_dict_[recommendation_type]
                        self.drop_columns.extend(columns_to_drop)
                        if update_transformer_log:
                            self.transformer_log(
                                utils.wrap_lines(
                                    columns_to_drop, heading="Drop constant columns:"
                                )
                            )
                    elif recommendation_type == "positive_class":
                        value = self.reco_dict_[recommendation_type][self.target_][
                            "Selected Action"
                        ]
                        if value != "Do nothing":
                            if self.target_ in X.columns:
                                # X = X.set_positive_class(value, missing_value=False)
                                X[self.target_] = X[self.target_].map(
                                    lambda x: x == value
                                )
                                if update_transformer_log:
                                    self.transformer_log(
                                        "Set %s as positive class" % value
                                    )
                    elif recommendation_type == "fix_imbalance" and self.fix_imbalance:
                        self.balancing_strategy = self.reco_dict_[recommendation_type][
                            self.target_
                        ]["Selected Action"]
                        if update_transformer_log:
                            self.transformer_log(
                                "Fix imbalance using technique: %s"
                                % self.balancing_strategy
                            )
                    else:
                        # Get the new column name if it has been combined with another
                        for column in self.reco_dict_[recommendation_type]:
                            selected_action = self.reco_dict_[recommendation_type][
                                column
                            ]["Selected Action"]

                            if selected_action.startswith("Drop"):
                                if selected_action != "Drop":
                                    column = selected_action.split(" ", 1)[1]
                                if column not in columns_to_drop:
                                    self.drop_columns.append(column)
                                    if update_transformer_log:
                                        self.transformer_log(
                                            'Drop: "{}"'.format(column)
                                        )
                            # elif selected_action.startswith('Combine'):
                            #     column1 = column
                            #     column2 = selected_action.split(" ", 2)[2]

                            #     if column1 not in self.drop_columns and column2 not in self.drop_columns:
                            #         self.combine_columns.append((column1, column2))
                            #         if update_transformer_log:
                            #             self.transformer_log('Combine: "{}" with "{}"'.format(column1, column2))
                            elif recommendation_type == "imputation":
                                fill_val = (
                                    self.fill_nan_dict_[column]
                                    if selected_action
                                    == "Fill missing values with constant"
                                    else get_fill_val(
                                        self.feature_metadata_,
                                        column,
                                        selected_action,
                                        constant="constant",
                                    )
                                )
                                if fill_val is not None:
                                    if column == self.target_:
                                        # target fill need not be reproduced at the time of scoring,
                                        # as it won't be present
                                        self.fill_na_target_ = fill_val
                                    else:
                                        if update_transformer_log:
                                            self.transformer_log(
                                                '{} in {}: "{}"'.format(
                                                    selected_action, column, fill_val
                                                )
                                            )
                                        self.fill_nan_dict_[column] = fill_val

        # Drop columns
        if len(self.drop_columns) > 0:
            logger.info("Dropping columns " + str(set(self.drop_columns)))
        X = X.drop(self.drop_columns, axis=1)

        # fill na
        if len(self.fill_nan_dict_) > 0:
            logger.info(
                "Filling NaN values in " + str(list(self.fill_nan_dict_.keys()))
            )
        for col, fill_val in self.fill_nan_dict_.items():
            if col in X:
                if (
                    X[col].dtype.name == "category"
                    and fill_val not in X[col].cat.categories.tolist()
                ):
                    X[col] = X[col].cat.add_categories([fill_val])
                X[col] = X[col].fillna(fill_val)

        # fix imbalance only at the time of initial fit, the subsequent transform calls are used to
        # reproduce transformations, and sampling is not required at that time
        if fit_transform and self.fix_imbalance:
            if self.balancing_strategy and self.balancing_strategy != "Do nothing":
                progress.update("Fixing imbalance by %s" % self.balancing_strategy)
                # The imputation during transformation uses the sampled df to find the columns contains nan values.
                # This strategy could miss  some columns with very few nans in the larger df. The up_sample
                # takes care of filling out such missed nan values. Downsample is not affected
                X = (
                    up_sample(X, self.target_, feature_types=self.feature_metadata_)
                    if self.balancing_strategy == "Up-sample"
                    else down_sample(X, self.target_)
                )
            else:
                progress.update()
        else:
            progress.update()
        return X

    def _get_na_action(self, recommendations, column, null_counts, null_ratio):
        if isinstance(self.feature_metadata_[column], ContinuousTypedFeature):
            if null_ratio == 1:
                possible_actions = [
                    "Drop",
                    "Fill missing values with constant",
                    "Do nothing",
                ]
                selected_action = "Drop"
            else:
                possible_actions = [
                    "Drop",
                    "Fill missing values with mean",
                    "Fill missing values with median",
                    "Fill missing values with frequent",
                    "Fill missing values with constant",
                    "Do nothing",
                ]
                if null_ratio <= 0.4:
                    selected_action = "Fill missing values with mean"
                else:
                    selected_action = "Drop"
        else:
            if null_ratio == 1:
                possible_actions = [
                    "Drop",
                    "Fill missing values with constant",
                    "Do nothing",
                ]
                selected_action = "Drop"
            else:
                possible_actions = [
                    "Drop",
                    "Fill missing values with frequent",
                    "Fill missing values with constant",
                    "Do nothing",
                ]
                if null_ratio <= 0.4:
                    selected_action = "Fill missing values with frequent"
                else:
                    selected_action = "Drop"
        if null_ratio < 0.1:
            msg = "Contains missing values({0})".format(null_counts)
        else:
            msg = "Contains missing values({0:.2%})".format(null_ratio)
        self._build_recommendation(
            recommendations,
            "imputation",
            [column],
            msg,
            possible_actions,
            selected_action,
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import uuid

from sklearn.pipeline import Pipeline

from ads.dataset import helper


class TransformerPipeline(Pipeline):
    def __init__(self, steps):
        assert len(steps) > 0, "steps is empty"
        self.step_names = set()
        steps = [self._get_step(step) for step in steps]
        super(TransformerPipeline, self).__init__(steps=steps)

    def add(self, transformer):
        """
        Add transformer to data transformation pipeline

        Parameters
        ----------
        transformer: Union[TransformerMixin, tuple(str, TransformerMixin)]
               if tuple, (name, transformer implementing transform)

        """
        step = self._get_step(transformer)
        self.steps.append(step)

    def _get_step(self, step):
        """
        Generate unique step name and transformer to add in pipeline.

        Parameters
        ----------
        step: transformer instance implementing fit and transform methods

        Retuns
        ------
        (unique_step_name, transformer) : tuple
        """
        if isinstance(step, tuple):
            name = step[0]
            step = step[1]
        else:
            name = step.__class__.__name__
            step = step
        if name in self.step_names:
            name = name + "/" + str(uuid.uuid4())
        self.step_names.add(name)
        return name, step

    def visualize(self):
        helper.visualize_transformation(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from typing import Any, Callable, Dict, List, Optional, Union, Tuple

import numpy as np
import pandas as pd
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.model.extractor.spark_extractor import SparkExtractor
from ads.model.serde.model_input import SparkModelInputSerializerType
from ads.model.serde.model_serializer import SparkModelSerializerType
from ads.model.generic_model import (
    FrameworkSpecificModel,
    DEFAULT_MODEL_FOLDER_NAME,
)
from ads.model.model_properties import ModelProperties
from ads.model.serde.common import SERDE

SPARK_DATAFRAME_SCHEMA_PATH = "_input_data_schema.json"


class SparkPipelineModel(FrameworkSpecificModel):
    """SparkPipelineModel class for estimators from the pyspark framework.

    Attributes
    ----------
    algorithm: str
        The algorithm of the model.
    artifact_dir: str
        Artifact directory to store the files needed for deployment.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained pyspark estimator/model using pyspark.
    framework: str
        "spark", the framework name of the model.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model.
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
        For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples
    --------
    >>> import tempfile
    >>> from ads.model.framework.spark_model import SparkPipelineModel
    >>> from pyspark.ml.linalg import Vectors
    >>> from pyspark.ml.classification import LogisticRegression

    >>> training = spark.createDataFrame([
    >>>     (1.0, Vectors.dense([0.0, 1.1, 0.1])),
    >>>     (0.0, Vectors.dense([2.0, 1.0, -1.0])),
    >>>     (0.0, Vectors.dense([2.0, 1.3, 1.0])),
    >>>     (1.0, Vectors.dense([0.0, 1.2, -0.5]))], ["label", "features"])
    >>> lr_estimator = LogisticRegression(maxIter=10, regParam=0.001)
    >>> pipeline = Pipeline(stages=[lr_estimator])
    >>> pipeline_model = pipeline.fit(training)

    >>> spark_model = SparkPipelineModel(estimator=pipeline_model, artifact_dir=tempfile.mkdtemp())
    >>> spark_model.prepare(inference_conda_env="dataexpl_p37_cpu_v3")
    >>> spark_model.verify(training)
    >>> spark_model.save()
    >>> model_deployment = spark_model.deploy()
    >>> spark_model.predict(training)
    >>> spark_model.delete_deployment()
    """

    _PREFIX = "spark"
    model_input_serializer_type = SparkModelInputSerializerType
    model_save_serializer_type = SparkModelSerializerType

    @runtime_dependency(
        module="pyspark",
        short_name="ml",
        object="ml",
        install_from=OptionalDependency.SPARK,
    )
    def __init__(
        self,
        estimator: Callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = model_save_serializer_type.SPARK,
        model_input_serializer: Optional[SERDE] = model_input_serializer_type.SPARK,
        **kwargs,
    ):
        """
        Initiates a SparkPipelineModel instance.

        Parameters
        ----------
        estimator: Callable
            SparkPipelineModel
        artifact_dir: str
            Directory for generate artifact.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to `ads.model.serde.model_input.SparkModelInputSERDE`.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        SparkPipelineModel
            SparkPipelineModel instance.


        Examples
        --------
        >>> import tempfile
        >>> from ads.model.framework.spark_model import SparkPipelineModel
        >>> from pyspark.ml.linalg import Vectors
        >>> from pyspark.ml.classification import LogisticRegression
        >>> from pyspark.ml import Pipeline

        >>> training = spark.createDataFrame([
        >>>     (1.0, Vectors.dense([0.0, 1.1, 0.1])),
        >>>     (0.0, Vectors.dense([2.0, 1.0, -1.0])),
        >>>     (0.0, Vectors.dense([2.0, 1.3, 1.0])),
        >>>     (1.0, Vectors.dense([0.0, 1.2, -0.5]))], ["label", "features"])
        >>> lr_estimator = LogisticRegression(maxIter=10, regParam=0.001)
        >>> pipeline = Pipeline(stages=[lr_estimator])
        >>> pipeline_model = pipeline.fit(training)

        >>> spark_model = SparkPipelineModel(estimator=pipeline_model, artifact_dir=tempfile.mkdtemp())
        >>> spark_model.prepare(inference_conda_env="pyspark30_p37_cpu_v5")
        >>> spark_model.verify(training)
        >>> spark_model.save()
        >>> model_deployment = spark_model.deploy()
        >>> spark_model.predict(training)
        >>> spark_model.delete_deployment()
        """
        if not (type(estimator) in [ml.PipelineModel]):
            raise TypeError(
                f"{str(type(estimator))} is not supported in `SparkPipelineModel`s."
            )
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = SparkExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter

    @staticmethod
    def _handle_model_file_name(as_onnx: bool, model_file_name: str):
        """
        Process folder name for saving model.

        Parameters
        ----------
        as_onnx: bool
            To convert to onnx format
        model_file_name: str
            File name for saving model.

        Returns
        -------
        str
            Processed file name. (Folder in the case of spark serialization)
        """
        if as_onnx:
            raise NotImplementedError(
                "The Spark to Onnx Conversion is not supported because it is unstable. Please set as_onnx to False (default) to perform a spark model serialization"
            )
        if not model_file_name:
            return DEFAULT_MODEL_FOLDER_NAME
        return model_file_name

    def serialize_model(
        self,
        as_onnx: bool = False,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
                "pyspark.sql.DataFrame",
                "pyspark.pandas.DataFrame",
            ]
        ] = None,
        force_overwrite: bool = False,
        **kwargs,
    ) -> None:
        """
        Serialize and save pyspark model using spark serialization.

        Parameters
        ----------
        force_overwrite: (bool, optional). Defaults to False.
            If set as True, overwrite serialized model if exists.

        Returns
        -------
        None
        """
        if as_onnx:
            raise NotImplementedError(
                "The Spark to Onnx Conversion is not supported because it is unstable. Please set as_onnx to False (default) to perform a spark model serialization"
            )

        super().serialize_model(
            as_onnx=as_onnx,
            force_overwrite=force_overwrite,
            X_sample=X_sample,
            **kwargs,
        )

    @runtime_dependency(
        module="pyspark",
        short_name="sql",
        object="sql",
        install_from=OptionalDependency.SPARK,
    )
    def _prepare_data_for_schema(
        self,
        X_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        **kwargs,
    ):
        """Generate Spark Schema and format Spark DataFrame as Pandas for ADS Schema Generation"""
        input_schema_path = os.path.join(
            self.artifact_dir,
            self.model_file_name + SPARK_DATAFRAME_SCHEMA_PATH,
        )
        (
            X_sample,
            data_type,
            schema,
        ) = self.get_data_serializer()._serialize_via_spark(X_sample)
        if not schema:
            raise TypeError(
                f"Data type: {data_type} unsupported. Please use `pyspark.sql.DataFrame`, `pyspark.pandas.DataFrame`, or `pandas.DataFrame`."
            )
        with open(input_schema_path, "w") as f:
            f.write(schema.json())

        if isinstance(X_sample, sql.DataFrame):
            X_sample = X_sample.toPandas()

        return X_sample, y_sample



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module helping to load/save images from/to the local path or OCI object storage bucket.

Classes
-------
ADSImage
    Work with image files that are stored in Oracle Cloud Infrastructure Object Storage.

Examples
--------
>>> from ads.feature_engineering import ADSImage
>>> from IPython.core.display import display
>>> img = ADSImage.open("1.jpg")
>>> display(img)
>>> img.save("oci://<bucket_name>@<namespace>/1.jpg")
>>> img1 = ADSImage.open("oci://<bucket_name>@<namespace>/1.jpg")
>>> display(img1)
"""

import os
from io import BytesIO
from typing import Dict, Optional

import fsspec
from ads.common import auth as authutil
from PIL import Image


class ADSImage:
    """
    Work with image files that are stored in Oracle Cloud Infrastructure Object Storage.
    PIL (Python Imaging Library) is used as a backend to represent and manipulate images.
    The PIL adds support
    for opening, manipulating, processing and saving various image file formats.

    Attributes
    ----------
    img: Image.Image
        A PIL Image object.
    filename: str
        The image filename.

    Methods
    -------
    save(self, path: str, ...) -> None
        Saves the image under the given filename.
    open(cls, path: str, ...) -> ADSImage
        Opens the given image file.

    Examples
    --------
    >>> from ads.feature_engineering import ADSImage
    >>> from IPython.core.display import display
    >>> img = ADSImage.open("1.jpg")
    >>> img.save("oci://<bucket_name>@<namespace>/1.jpg")
    >>> img1 = ADSImage.open("oci://<bucket_name>@<namespace>/1.jpg")
    >>> display(img1)
    """

    def __init__(self, img: Image.Image, filename: Optional[str] = None) -> None:
        """Initializes ADSImage object.

        Parameters
        ----------
        img: PIL.Image.Image
            The PIL Image object.
        filename: (str, optional). Defaults to None.
            The image filename.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            If `img` is not an instance of `PIL.Image.Image`.
        ValueError
            If `img` is not provided.
        """
        if not img:
            raise ValueError("The parameter `img` is required.")

        if not isinstance(img, Image.Image):
            raise TypeError("The `img` parameter must be a `PIL.Image.Image` object.")

        self.img = img
        self.filename = filename

    def save(
        self,
        path: str,
        format: Optional[str] = None,
        auth: Optional[Dict] = None,
        **kwargs: Optional[Dict]
    ) -> None:
        """Save the image under the given filename.
        If no format is specified, the format to use is determined from the image object
        or filename extension, if possible.

        Parameters
        ----------
        path: str
            The file path to save image. It can be a local path or an Oracle Cloud Infrastructure Object Storage URI.
            Example: `oci://<bucket_name>@<namespace>/1.jpg`
        format: (str, optional). Defaults to None.
            If omitted and `path` has a file extension the format of the image will be based on the extension.
            Can be any format supported by PIL Image.
            The available options are described in the PIL image format documentation:
            https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html

        auth: (Dict, optional). Defaults to None.
            The default authentication is set using `ads.set_auth()` API. To override the
            default behavior, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create an
            authentication signer and provide an `IdentityClient` object.
        kwargs:
            Additional keyword arguments that would be passed to the PIL Image `save()` method.

        Returns
        -------
        None
            Nothing.
        """
        imgByteArr = BytesIO()
        self.img.save(imgByteArr, format=format or self.img.format, params=kwargs)
        auth = auth or authutil.default_signer()
        with fsspec.open(path, mode="wb", **auth) as f:
            f.write(imgByteArr.getvalue())

    @classmethod
    def open(cls, path: str, storage_options: Optional[Dict] = None) -> "ADSImage":
        """Opens the given image file.

        Parameters
        ----------
        path: str
            The file path to open image. Can be local path or OCI object storage URI.
            Example: `oci://<bucket_name>@<namespace>/1.jpg`
        storage_options: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        ADSImage
            The instance of `ADSImage`.
        """
        if storage_options == None:
            storage_options = authutil.default_signer()
        with fsspec.open(path, mode="rb", **storage_options) as f:
            return cls(
                img=Image.open(BytesIO(f.read())),
                filename=os.path.basename(f.path),
            )

    def to_bytes(self, **kwargs: Optional[Dict]) -> BytesIO:
        """Converts image to bytes.
        If no format is specified, the format to use is determined from the image object if it possible.

        Parameters
        ---------------
        kwargs:
            format: (str, optional). Defaults to None.
                If omitted and `path` has a file extension the format of the image will be based on the extension.
                Can be any format supported by PIL Image.
                The available options are described in the PIL image format documentation:
                https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
            Additional keyword arguments that would be passed to the PIL Image `save()` method.

        Returns
        -------
        BytesIO
            The image converted to bytes.
        """
        imgByteArr = BytesIO()
        self.img.save(
            imgByteArr, format=kwargs.get("format") or self.img.format, params=kwargs
        )
        return imgByteArr.getvalue()

    def __getattr__(self, key: str):
        if key == "img":
            raise AttributeError()
        return getattr(self.img, key)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict

from ads.common.object_storage_details import ObjectStorageDetails
from ads.common.serializer import DataClassSerializable
from ads.config import CONDA_BUCKET_NAME, CONDA_BUCKET_NS
from ads.model.runtime.utils import (
    INFERENCE_ENV_SCHEMA_PATH,
    TRAINING_ENV_SCHEMA_PATH,
    SchemaValidator,
    get_service_packs,
)


DEFAULT_CONDA_BUCKET_NAME = "service-conda-packs"


class PACK_TYPE(Enum):
    """Conda Pack Type"""

    SERVICE_PACK = "data_science"
    USER_CUSTOM_PACK = "published"


class EnvInfo(ABC):
    """Env Info Base class."""

    @classmethod
    @abstractmethod
    def _populate_env_info(
        cls, env_slug: str, env_type: str, env_path: str, python_version: str
    ) -> "EnvInfo":
        """Populate the EnvInfo instance.

        Parameters
        ----------
        env_slug: (str)
            conda pack slug.
        env_type: (str)
            conda pack type: data_science or published.
        env_path: (str)
            conda pack object storage path.
        python_version: (str)
            python version of the conda pack.

        Returns
        -------
        EnvInfo
            An EnvInfo instance.
        """
        pass

    @classmethod
    def from_slug(
        cls,
        env_slug: str,
        namespace: str = CONDA_BUCKET_NS,
        bucketname: str = CONDA_BUCKET_NAME,
        auth: dict = None,
    ) -> "EnvInfo":
        """Initiate an EnvInfo object from a slug. Only service pack is allowed to use this method.

        Parameters
        ----------
        env_slug: str
            service pack slug.
        namespace: (str, optional)
            namespace of region.
        bucketname: (str, optional)
            bucketname of service pack.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        EnvInfo
            An EnvInfo instance.
        """
        if "oci://" not in env_slug:
            warnings.warn("slug will be deprecated. Provide conda pack path instead.")

        if not bucketname:
            warnings.warn(
                f"`bucketname` is not provided, defaults to `{DEFAULT_CONDA_BUCKET_NAME}`."
            )
            bucketname = DEFAULT_CONDA_BUCKET_NAME
        _, service_pack_slug_mapping = get_service_packs(
            namespace, bucketname, auth=auth
        )
        env_type, env_path, python_version = None, None, None
        if service_pack_slug_mapping:
            if env_slug in service_pack_slug_mapping:
                env_type = PACK_TYPE.SERVICE_PACK.value
                env_path, python_version = service_pack_slug_mapping[env_slug]
            else:
                warnings.warn(
                    "The {env_slug} is not a service pack. Use `from_path` method by passing in the object storage path."
                )

        return cls._populate_env_info(
            env_slug=env_slug,
            env_type=env_type,
            env_path=env_path,
            python_version=python_version,
        )

    @classmethod
    def from_path(cls, env_path: str, auth: dict = None) -> "EnvInfo":
        """Initiate an object from a conda pack path.

        Parameters
        ----------
        env_path: str
            conda pack path.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        EnvInfo
            An EnvInfo instance.
        """
        bucketname, namespace, _ = ObjectStorageDetails.from_path(env_path).to_tuple()
        env_type = ""
        python_version = ""
        env_slug = ""
        service_pack_path_mapping = {}
        service_pack_path_mapping, _ = get_service_packs(
            namespace, bucketname, auth=auth
        )
        if env_path.startswith("oci://") and service_pack_path_mapping:
            if env_path in service_pack_path_mapping:
                env_type = PACK_TYPE.SERVICE_PACK.value
                (
                    env_slug,
                    python_version,
                ) = service_pack_path_mapping[env_path]
            else:
                env_type = PACK_TYPE.USER_CUSTOM_PACK.value
                try:
                    metadata_json = ObjectStorageDetails.from_path(
                        env_path
                    ).fetch_metadata_of_object()
                    python_version = metadata_json.get("python", None)
                    env_slug = metadata_json.get("slug", None)
                except Exception as e:
                    logging.warning(e)
                    logging.warning(
                        "python version and slug are not found from the manifest metadata."
                    )

        return cls._populate_env_info(
            env_slug=env_slug,
            env_type=env_type,
            env_path=env_path,
            python_version=python_version,
        )

    @staticmethod
    def _validate(obj_dict: Dict, schema_file_path: str) -> bool:
        """Validate the content in the ditionary format from the yaml file.

        Parameters
        ----------
        obj_dict: (Dict)
            yaml file content to validate.

        Returns
        -------
        bool
            Validation result.
        """
        validator = SchemaValidator(schema_file_path=schema_file_path)
        return validator.validate(document=obj_dict)


@dataclass(repr=False)
class TrainingEnvInfo(EnvInfo, DataClassSerializable):
    """Training conda environment info."""

    training_env_slug: str = ""
    training_env_type: str = ""
    training_env_path: str = ""
    training_python_version: str = ""

    @classmethod
    def _populate_env_info(
        cls, env_slug: str, env_type: str, env_path: str, python_version: str
    ) -> "TrainingEnvInfo":
        """Populate the TrainingEnvInfo instance.

        Parameters
        ----------
        env_slug: (str)
            conda pack slug.
        env_type: (str)
            conda pack type: data_science or published.
        env_path: (str)
            conda pack object storage path.
        python_version: (str)
            python version of the conda pack.

        Returns
        -------
        TrainingEnvInfo
            An TrainingEnvInfo instance.
        """
        return cls(
            training_env_slug=env_slug,
            training_env_type=env_type,
            training_env_path=env_path,
            training_python_version=python_version,
        )

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """Validate the content in the ditionary format from the yaml file.

        Parameters
        ----------
        obj_dict: (Dict)
            yaml file content to validate.

        Returns
        -------
        bool
            Validation result.
        """
        return EnvInfo._validate(
            obj_dict=obj_dict, schema_file_path=TRAINING_ENV_SCHEMA_PATH
        )


@dataclass(repr=False)
class InferenceEnvInfo(EnvInfo, DataClassSerializable):
    """Inference conda environment info."""

    inference_env_slug: str = ""
    inference_env_type: str = ""
    inference_env_path: str = ""
    inference_python_version: str = ""

    @classmethod
    def _populate_env_info(
        cls, env_slug: str, env_type: str, env_path: str, python_version: str
    ) -> "InferenceEnvInfo":
        """Populate the InferenceEnvInfo instance.

        Parameters
        ----------
        env_slug: (str)
            conda pack slug.
        env_type: (str)
            conda pack type: data_science or published.
        env_path: (str)
            conda pack object storage path.
        python_version: (str)
            python version of the conda pack.

        Returns
        -------
        InferenceEnvInfo
            An InferenceEnvInfo instance.
        """
        return cls(
            inference_env_slug=env_slug,
            inference_env_type=env_type,
            inference_env_path=env_path,
            inference_python_version=python_version,
        )

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """Validate the content in the ditionary format from the yaml file.

        Parameters
        ----------
        obj_dict: (Dict)
            yaml file content to validate.

        Returns
        -------
        bool
            Validation result.
        """
        return EnvInfo._validate(
            obj_dict=obj_dict, schema_file_path=INFERENCE_ENV_SCHEMA_PATH
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC, abstractmethod
from typing import Any
from ads.common.serializer import Serializable


class Reader(ABC):
    """Data Reader Interface."""

    def info(self) -> Serializable:
        NotImplementedError(
            f"The class {self.__class__.__name__} did not implement the required method "
            "`info()`. Contact the class maintainer."
        )

    @abstractmethod
    def read(self) -> Any:
        pass



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from string import Template
from typing import Dict
import json

import yaml

from ads.common.auth import AuthType
from ads.opctl import logger
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.utils import read_from_ini, _DefaultNoneDict
from ads.opctl.utils import is_in_notebook_session, get_service_pack_prefix
from ads.opctl.constants import (
    DEFAULT_PROFILE,
    DEFAULT_OCI_CONFIG_FILE,
    DEFAULT_CONDA_PACK_FOLDER,
    DEFAULT_ADS_CONFIG_FOLDER,
    ADS_JOBS_CONFIG_FILE_NAME,
    ADS_CONFIG_FILE_NAME,
    ADS_ML_PIPELINE_CONFIG_FILE_NAME,
    ADS_DATAFLOW_CONFIG_FILE_NAME,
    ADS_LOCAL_BACKEND_CONFIG_FILE_NAME,
    DEFAULT_NOTEBOOK_SESSION_CONDA_DIR,
    BACKEND_NAME,
)


class ConfigMerger(ConfigProcessor):
    """Merge configurations from command line args, YAML, ini and default configs.
    The order of precedence is
    command line args > YAML + conf.ini > .ads_ops/config.ini + .ads_ops/ml_job_config.ini or .ads_ops/dataflow_config.ini
    Detailed examples can be found at the last section on this page:
    https://bitbucket.oci.oraclecorp.com/projects/ODSC/repos/advanced-ds/browse/ads/opctl/README.md?at=refs%2Fheads%2Fads-ops-draft
    """

    def process(self, **kwargs) -> None:
        config_string = Template(json.dumps(self.config)).safe_substitute(os.environ)
        self.config = json.loads(config_string)
        # 1. merge and overwrite values from command line args
        self._merge_config_with_cmd_args(kwargs)
        # 1.5 merge environment variables
        # TODO

        # 2. fill in values from conf file
        self._fill_config_from_conf()

        ads_config_path = os.path.abspath(
            os.path.expanduser(
                self.config["execution"].pop("ads_config", DEFAULT_ADS_CONFIG_FOLDER)
            )
        )
        # 3. fill in values from default files under ~/.ads_ops
        self._fill_config_with_defaults(ads_config_path)

        logger.debug(f"Config: {self.config}")
        return self

    def _merge_config_with_cmd_args(self, cmd_args: Dict) -> None:
        # overwrite config with command line args
        # if a command line arg value is None or empty collection, then it is ignored
        def _overwrite(cfg, args):
            for k, v in cfg.items():
                if isinstance(v, dict):
                    _overwrite(v, args)
                elif k in args:
                    if (
                        isinstance(args[k], bool)
                        or (isinstance(args[k], list) and len(args[k]) > 0)
                        or args[k]
                    ):
                        cfg[k] = args.pop(k)
                    else:
                        args.pop(k)

        _overwrite(self.config, cmd_args)

        # save everything else from command line in "execution" section
        if "execution" not in self.config:
            self.config["execution"] = {}
        for k, v in cmd_args.items():
            if isinstance(v, bool) or (isinstance(v, list) and len(v) > 0) or v:
                self.config["execution"][k] = v

    def _fill_config_from_conf(self) -> None:
        if self.config["execution"].get("conf_file"):
            conf_file = self.config["execution"]["conf_file"]
            conf_profile = self.config["execution"].get("conf_profile", DEFAULT_PROFILE)
            logger.info(f"Reading from {conf_profile} using profile {conf_profile}")
            parser = read_from_ini(conf_file)
            extra_configs = dict(os.environ)
            extra_configs.update(parser[conf_profile])
            config_string = yaml.dump(self.config)
            # _DefaultNoneDict is used so that if $variable is not found in a section in .ini, None value is filled in.
            self.config = yaml.safe_load(
                Template(config_string).substitute(_DefaultNoneDict(**extra_configs))
            )

    def _fill_config_with_defaults(self, ads_config_path: str) -> None:
        exec_config = self._get_default_config()
        exec_config_from_conf = self._get_config_from_config_ini(ads_config_path)
        exec_config.update(
            {k: v for k, v in exec_config_from_conf.items() if v is not None}
        )
        # set default auth
        if not self.config["execution"].get("auth", None):
            if is_in_notebook_session():
                self.config["execution"]["auth"] = AuthType.RESOURCE_PRINCIPAL
            else:
                self.config["execution"]["auth"] = AuthType.API_KEY
        # determine profile
        if self.config["execution"]["auth"] != AuthType.API_KEY:
            profile = self.config["execution"]["auth"].upper()
            exec_config.pop("oci_profile", None)
            self.config["execution"]["oci_profile"] = None
        else:
            profile = self.config["execution"].get("oci_profile") or exec_config.get(
                "oci_profile"
            )
            self.config["execution"]["oci_profile"] = profile
        # loading config for corresponding profile
        logger.info(f"Loading service config for profile {profile}.")
        infra_config = self._get_service_config(profile, ads_config_path)
        if infra_config.get(
            "conda_pack_os_prefix"
        ):  # this is a field that appeared both in config.ini and ml_job_config.ini
            exec_config["conda_pack_os_prefix"] = infra_config.pop(
                "conda_pack_os_prefix"
            )
        for k, v in exec_config.items():
            if v and not self.config["execution"].get(k):
                self.config["execution"][k] = v
        if not self.config.get("infrastructure"):
            self.config["infrastructure"] = {}
        for k, v in infra_config.items():
            if v and not self.config["infrastructure"].get(k):
                self.config["infrastructure"][k] = v

    def _get_default_config(self) -> Dict:
        if is_in_notebook_session():
            conda_pack_os_prefix = get_service_pack_prefix()
            return {
                "oci_config": DEFAULT_OCI_CONFIG_FILE,
                "oci_profile": DEFAULT_PROFILE,
                "conda_pack_folder": DEFAULT_NOTEBOOK_SESSION_CONDA_DIR,
                "conda_pack_os_prefix": conda_pack_os_prefix,
            }
        else:
            return {
                "oci_config": DEFAULT_OCI_CONFIG_FILE,
                "oci_profile": DEFAULT_PROFILE,
                "conda_pack_folder": DEFAULT_CONDA_PACK_FOLDER,
            }

    @staticmethod
    def _get_config_from_config_ini(ads_config_folder: str) -> Dict:
        if os.path.exists(os.path.join(ads_config_folder, ADS_CONFIG_FILE_NAME)):
            parser = read_from_ini(
                os.path.join(ads_config_folder, ADS_CONFIG_FILE_NAME)
            )
            return {
                "oci_config": parser["OCI"].get("oci_config"),
                "oci_profile": parser["OCI"].get("oci_profile"),
                "conda_pack_folder": parser["CONDA"].get("conda_pack_folder"),
                "conda_pack_os_prefix": parser["CONDA"].get("conda_pack_os_prefix"),
            }
        else:
            logger.info(
                f"{os.path.join(ads_config_folder, 'config.ini')} does not exist. No config loaded."
            )
            return {}

    def _get_service_config(self, oci_profile: str, ads_config_folder: str) -> Dict:
        backend = self.config["execution"].get("backend", None)
        backend_config = {
            BACKEND_NAME.DATAFLOW.value: ADS_DATAFLOW_CONFIG_FILE_NAME,
            BACKEND_NAME.PIPELINE.value: ADS_ML_PIPELINE_CONFIG_FILE_NAME,
            BACKEND_NAME.LOCAL.value: ADS_LOCAL_BACKEND_CONFIG_FILE_NAME,
        }
        config_file = backend_config.get(backend, ADS_JOBS_CONFIG_FILE_NAME)

        if os.path.exists(os.path.join(ads_config_folder, config_file)):
            parser = read_from_ini(os.path.join(ads_config_folder, config_file))
            if oci_profile in parser:
                return parser[oci_profile]
        else:
            logger.info(
                f"{os.path.join(ads_config_folder, config_file)} does not exist. No config loaded."
            )
        return {}



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import oci.data_science
from ads.common.oci_mixin import OCIModelMixin


class OCIDataScienceMixin(OCIModelMixin):
    @classmethod
    def init_client(cls, **kwargs) -> oci.data_science.DataScienceClient:
        return cls._init_client(client=oci.data_science.DataScienceClient, **kwargs)

    @property
    def client(self) -> oci.data_science.DataScienceClient:
        return super().client

    @property
    def client_composite(self) -> oci.data_science.DataScienceClientCompositeOperations:
        return oci.data_science.DataScienceClientCompositeOperations(self.client)


class DSCNotebookSession(OCIDataScienceMixin, oci.data_science.models.NotebookSession):
    """Represents a data science notebook session

    To get the information of an existing notebook session:
    >>> notebook = DSCNotebookSession.from_ocid(NOTEBOOK_OCID)
    Get the name of the notebook session
    >>> notebook.display_name
    Get the subnet ID of the notebook session
    >>> notebook.notebook_session_configuration_details.subnet_id
    """



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.opctl.config.base import ConfigProcessor


class ConfigValidator(ConfigProcessor):
    def process(self):
        # TODO: add validation using pydantic + datamodel-code-generator in future PR

        # called to update command, part of which is encoded spec
        # and spec might have been updated during the call above
        # self.["execution"]["command"] = _resolve_command(self.config)
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import

import re, pathlib, os
import urllib.parse
from abc import ABC, abstractmethod
from os import listdir
from os.path import isfile, isdir, join, getsize
from typing import List, Set, Tuple, Dict

import requests

import pandas as pd
import sklearn.datasets as sk_datasets

from ads.dataset.factory import DatasetFactory
from ads.common.utils import inject_and_copy_kwargs
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class DatasetBrowser(ABC):
    #
    # anything that lists & loads datasets
    #

    def __init__(self):  # pragma: no cover
        pass

    @staticmethod
    def list(filter_pattern="*") -> List[str]:
        """
        Return a list of dataset browser strings.
        """
        return ["web", "sklearn", "seaborn", "GitHub"]

    @abstractmethod
    def open(name: str, **kwargs):  # pragma: no cover
        """
        Return new dataset for the given name.

        Parameters
        ----------
        name : str
            the name of the dataset to open.

        Returns
        -------
        ds: Dataset

        Examples
        --------
        ds_browser = DatasetBrowser("sklearn")

        ds = ds_browser.open("iris")
        """
        pass

    #
    # helper to filter list of dataset names
    #
    def filter_list(self, L, filter_pattern) -> List[str]:  # pragma: no cover
        """
        Filters a list of dataset names.
        """
        return list(filter(re.compile(filter_pattern).match, L))

    #
    #
    # Create subclasses for each type of DatasetBrowser
    #
    #

    @staticmethod
    def web(index_url: str):
        """
        Returns a WebDataset.
        """
        return WebDatasets(index_url)  # pragma: no cover

    @staticmethod
    def sklearn():
        """
        Returns a SklearnDataset.
        """
        return SklearnDatasets()

    @staticmethod
    def seaborn():
        """
        Returns a SeabornDataset.
        """
        return SeabornDatasets()

    @staticmethod
    def filesystem(folder: str):
        """
        Returns a LocalFilesystemDataset.
        """
        return LocalFilesystemDatasets(folder)

    @staticmethod
    def GitHub(user: str, repo: str, branch: str = "master"):
        """
        Returns a GitHubDataset
        """
        return GitHubDatasets(user, repo, branch)


class GitHubDatasets(DatasetBrowser):
    def __init__(self, user: str, repo: str, branch: str):
        super(DatasetBrowser, self).__init__()
        self.user = user
        self.repo = repo
        self.branch = branch
        self.api_url = (
            f"https://api.github.com/repos/{user}/{repo}/git/trees/{branch}?recursive=1"
        )

    def _generate_filelist(self):

        response = requests.get(self.api_url)
        if response.status_code == 200:
            listing = []
            j = response.json()
            if "tree" in j:
                for n in j["tree"]:
                    filepath = n["path"]
                    fileurl = f"https://raw.githubusercontent.com/{self.user}/{self.repo}/{self.branch}/{filepath}"
                    pl = pathlib.Path(filepath)
                    format = pl.suffix[1:]
                    if format.lower() in ["csv", "tsv", "json"]:
                        listing.append(
                            {
                                "url": fileurl,
                                "name": os.path.splitext(filepath)[0],
                                "format": format,
                                "size": n["size"],
                                "description": f"""
                                                Origin: GitHub {self.user}@{self.repo}#{self.branch}\n
                                                Name: {filepath}, size: {n['size']/1024}kb\n"
                                                """.strip(),
                            }
                        )

            return listing

        else:
            raise ValueError(
                f"Error accessing GitHub API ({self.api_url}): {response.status_code}"
            )

    def list(self, filter_pattern: str = ".*") -> List[str]:
        return super().filter_list(
            [x["name"] for x in self._generate_filelist()], filter_pattern
        )

    def open(self, name: str, **kwargs):
        #
        # lookup the name
        #

        for obj in self._generate_filelist():
            if obj["name"] == name:
                return DatasetFactory.open(
                    **inject_and_copy_kwargs(
                        kwargs,
                        **{
                            "source": obj["url"],
                            "format": obj["format"],
                            "name": obj["name"],
                            "description": obj["description"],
                        },
                    )
                )

        raise ValueError(f"dataset [{name}] does not exist, use .list() to display all")


class LocalFilesystemDatasets(DatasetBrowser):
    def __init__(self, folder: str):
        super(DatasetBrowser, self).__init__()
        self.folder = folder

    def list(self, filter_pattern: str = ".*") -> List[str]:
        return super().filter_list(
            [x["name"] for x in self._generate_filelist()], filter_pattern
        )

    def open(self, name: str, **kwargs):
        #
        # lookup the name
        #

        for obj in self._generate_filelist():
            if obj["name"] == name:
                return DatasetFactory.open(
                    **inject_and_copy_kwargs(
                        kwargs,
                        **{
                            "source": obj["path"],
                            "format": obj["format"],
                            "name": obj["name"],
                            "description": obj["description"],
                        },
                    )
                )

        raise ValueError(f"dataset [{name}] does not exist, use .list() to display all")

    def _generate_filelist(self):
        if not isdir(self.folder):
            raise ValueError(
                f'The path "{self.folder}" does not exist, or is not a folder'
            )
        else:
            onlyfiles = [
                f for f in listdir(self.folder) if isfile(join(self.folder, f))
            ]
            listing = []
            for f in onlyfiles:
                i = f.rfind(".")
                if i > 0:
                    name, format = f[0:i].strip(), f[i + 1 :].strip()
                    path = join(self.folder, f)
                    if format.lower() in ["json", "csv", "tsv", "hdf"]:
                        listing.append(
                            {
                                "path": path,
                                "size": getsize(path),
                                "name": f"{name}.{format}",
                                "format": format.strip(),
                                "description": f"""
                                                Origin: {self.folder}\n
                                                Name: {f} ({getsize(path)} bytes)\n"
                                                """.strip(),
                            }
                        )

            return listing


class WebDatasets(DatasetBrowser):
    @runtime_dependency(module="htmllistparse", install_from=OptionalDependency.DATA)
    def __init__(self, index_url: str):  # pragma: no cover

        self.index_url = index_url
        self.listing = []
        super(DatasetBrowser, self).__init__()

        try:
            _, raw_listing = htmllistparse.fetch_listing(self.index_url, timeout=30)
        except Exception as e:
            raise ValueError(str(e))

        for f in [x for x in raw_listing if x.size]:
            i = f.name.rfind(".")
            filename, format = f.name[0:i].strip(), f.name[i + 1 :]
            if format.lower() in ["json", "csv", "tsv", "hdf"]:
                d = {
                    "url": urllib.parse.urljoin(self.index_url, f.name),
                    "size": f.size,
                    "name": filename,
                    "format": format,
                    "description": f"""
                                        Origin: {self.index_url}\n
                                        Name: {f.name.strip()} ({f.size} bytes)\n"
                                        """.strip(),
                }
                self.listing.append(d)

    def list(self, filter_pattern: str = ".*") -> List[str]:
        return super().filter_list([x["name"] for x in self.listing], filter_pattern)

    def open(self, name: str, **kwargs):
        #
        # lookup the name
        #
        for obj in self.listing:
            if obj["name"] == name:
                return DatasetFactory.open(
                    obj["url"],
                    format=obj["format"],
                    name=obj["name"],
                    description=obj["description"],
                )
        raise ValueError(f"dataset [{name}] does not exist, use .list() to display all")


class SeabornDatasets(DatasetBrowser):
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def __init__(self):
        super(DatasetBrowser, self).__init__()
        self.dataset_names = list(seaborn.get_dataset_names())

    def list(self, filter_pattern: str = ".*") -> List[str]:
        return super().filter_list(self.dataset_names, filter_pattern)

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def open(self, name: str, **kwargs):
        if name in self.dataset_names:
            return DatasetFactory.open(
                seaborn.load_dataset(name), name=name, description="from seaborn"
            )
        else:
            raise ValueError(
                "dataset [{name}] does not exist, use .list() to display all"
            )


class SklearnDatasets(DatasetBrowser):

    sklearn_datasets = ["breast_cancer", "diabetes", "iris", "wine", "digits"]

    def __init__(self):
        super(DatasetBrowser, self).__init__()

    def list(self, filter_pattern: str = ".*") -> List[str]:
        return super().filter_list(SklearnDatasets.sklearn_datasets, filter_pattern)

    def open(self, name: str, **kwargs):
        if name in SklearnDatasets.sklearn_datasets:
            data = getattr(sk_datasets, "load_%s" % (name))()
            description = data["DESCR"]
            if "images" in data:
                # special case digits
                n_samples = len(data.images)
                cols = data.images.reshape((n_samples, -1))
                df = pd.DataFrame(
                    cols, columns=["f%d" % (i) for i in range(cols.shape[1])]
                )
                for col in df.columns:
                    df[col] = df[col].astype(float)
                df["target"] = pd.Series(data.target).astype("category")
            elif "target_names" in data:
                # inverse transform the target labels for categorical types
                df = pd.DataFrame(data.data, columns=data.feature_names)
                df["target"] = pd.Series(
                    [data.target_names[x] for x in data.target]
                ).astype("category")
            else:
                df = pd.DataFrame(data.data, columns=data.feature_names)
                df["target"] = pd.Series(data.target)

            return DatasetFactory.open(
                df, target="target", name=name, description=description
            )

        else:
            raise ValueError(
                f"dataset [{name}] does not exist, use .list() to display all"
            )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np

from ads.explanations.base_explainer import WhatIfExplainer
from ads.explanations.mlx_interface import init_whatif_explainer
from ads.common import logger
from ads.common.utils import _log_multivalue_feature_column_error
from ads.common.decorator.deprecate import deprecated


class MLXWhatIfExplainer(WhatIfExplainer):
    """
    MLX WhatIf class.

    Generates "what if" explanations to conduct sensitivity test. Supported explanations:

        - (Interactive Widgets and Plotly Visualization) Explore sample.
        - (Interactive Widgets and Plotly Visualization) Explore predictions.

    Supports:

        - Binary classification. (Tabular Data only)
        - Regression. (Tabular Data only)

    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self):
        super(MLXWhatIfExplainer, self).__init__()
        self.explainer = None

    def explore_sample(self, row_idx=None, features=None, max_features=32, **kwargs):
        """
        Parameters
        ----------
        row_idx: int, optional
            Row index of the sample to explore. Default value is `None`
        features: list of str or int (dataframe column names), optional
            Feature columns to explore. Default value is `None`.
        max_features: int, optional
            Maximum number of features for modification. Default value is 32
        """

        if self.explainer is None:
            self.explainer = init_whatif_explainer(
                explainer=self.explainer,
                est=self.est_,
                X_test=self.X_test_,
                y_test=self.y_test_,
                mode=self.mode_,
                class_names=self.class_names_
                if self.mode_ == "classification"
                else ["prediction"],
                train=self.X_train_,
                target_title=None,
                **kwargs,
            )

        kwargs.pop("width", None)
        if row_idx is not None and row_idx not in self.X_test.index:
            logger.info(f"`row_idx` {row_idx} is not in the index. Resetting to 0.")
            # print('row_idx (%s) not in the index, reset to 0.'% str(row_idx))
            row_idx = 0

            # Wrap in a list if a list is not provided

        if features is not None and not isinstance(features, list):
            features = [features]
        if features:
            feature_names = self.X_train.columns.tolist()

            # Fail if we were not provided valid feature names
            if not all(np.isin(features, feature_names)):
                logger.error(
                    f"One or more `features` {str(features)} does not exist in data. "
                )
                logger.info(f"Existing features: {feature_names}.")
                return

        return self.explainer.explore_sample(
            row_idx=row_idx,
            features=features,
            max_features=max_features,
            plot_type="pyplot",
            **kwargs,
        )

    def explore_predictions(
        self,
        x=None,
        y=None,
        label=None,
        plot_type="scatter",
        discretization=None,
        **kwargs,
    ):
        """
        Explore explanation predictions

        Parameters
        ----------
        x: str, optional
            Feature column on x-axis. The default is `None`.
        y: str, optional
            Feature column or model prediction column on the y-axis, by default it is the target.
        label: str or int, optional
            Target label or target class name to explore only for classification problems.
            The default is `None`.
        plot_type: str, optional
            Type of plot. For classification problems the valid options are 'scatter',
            'box', or 'bar'. For a regression problem, the valid options are 'scatter' or 'box'.
            The default is 'scatter'.
        discretization: str, optional
            Discretization method applies the x-axis if the feature x is continuous.
            The valid options are 'quartile', 'decile', or 'percentile'. The default is `None`.
        """

        if self.explainer is None:
            self.explainer = init_whatif_explainer(
                explainer=self.explainer,
                est=self.est_,
                X_test=self.X_test_,
                y_test=self.y_test_,
                mode=self.mode_,
                class_names=self.class_names_
                if self.mode_ == "classification"
                else ["prediction"],
                train=self.X_train_,
                target_title=None,
                **kwargs,
            )

        kwargs.pop("width", None)
        if isinstance(x, list) or isinstance(y, list):
            _log_multivalue_feature_column_error()
            return
        for feature in [x, y]:
            if feature is not None:
                if isinstance(feature, list):
                    _log_multivalue_feature_column_error()
                if not isinstance(feature, str):
                    feature = str(feature)

                # Convert to uppercase to be case-insensitive
                feature_upper = feature.upper()
                feature_names = np.char.upper(self.X_test.columns.tolist())

                # Fail if we were not provided valid feature names
                if not np.isin(feature_upper, feature_names):
                    logger.error(f"`{feature_upper}` does not exist in data. ")
                    logger.info(
                        f"Existing features: `{str(self.X_test.columns.tolist())}`."
                    )
                    return
        if self.class_names_ is not None:
            num_classes = len(self.class_names_)
            if isinstance(label, str) and label not in self.class_names_:
                logger.error(f"`label` must be one of {self.class_names_}.")
                return
            if isinstance(label, int) and label > num_classes - 1:
                logger.error(
                    f"`label` cannot exceed the value of {int(num_classes - 1)}."
                )
                return

        plot_type_options = ["scatter", "bar", "box"]
        if plot_type not in plot_type_options:
            logger.error(f"`plot_type` must be one of {str(plot_type_options)}.")
            return
        discretization_options = [None, "quartile", "decile", "percentile"]
        if discretization not in discretization_options:
            logger.error(
                f"`discretization` must be one of {str(discretization_options)}."
            )
            return

        return self.explainer.explore_predictions(
            x=x,
            y=y,
            label=label,
            plot_type=plot_type,
            discretization=discretization,
            **kwargs,
        )


class WhatIfExplanationsException(TypeError):
    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, msg):
        super(WhatIfExplanationsException, self).__init__(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import abc
from ads.model.model_metadata import MetadataTaxonomyKeys
from typing import Dict


def normalize_hyperparameter(data: Dict) -> dict:
    """Converts all the fields to string to make sure it's
    json serializable.

    Parameters
    ----------
    data: ([Dict])
        The hyperparameter returned by the model.

    Returns
    -------
    Dict
        Normalized (json serializable) dictionary.
    """
    if not data:
        return None
    new_dict = {}
    for key, value in data.items():
        if isinstance(value, dict):
            new_dict[key] = normalize_hyperparameter(value)
        else:
            try:
                new_dict[key] = str(value)
            except:
                new_dict[key] = ""
    return new_dict


class ModelInfoExtractor(abc.ABC):
    """The base abstract class to extract model metadata.

    Methods
    -------
    framework(self) -> str
        Returns the framework of the model.
    algorithm(self) -> object
        Returns the algorithm of the model.
    version(self) -> str
        Returns the version of framework of the model.
    hyperparameter(self) -> dict
        Returns the hyperparameter of the model.
    info(self) -> dict
        Returns the model taxonomy metadata information.
    """

    @abc.abstractmethod
    def framework(self):
        """The abstract method to extracts the framework of the model.

        Returns
        ----------
        str:
           The framework of the model.
        """
        pass

    @abc.abstractmethod
    def algorithm(self):
        """The abstract method to extracts the algorithm of the model.

        Returns
        ----------
        object:
           The algorithm of the model.
        """
        pass

    @abc.abstractmethod
    def version(self):
        """The abstract method to extracts the framework version of the model.

        Returns
        ----------
        str:
           The framework version of the model.
        """
        pass

    @abc.abstractmethod
    def hyperparameter(self):
        """The abstract method to extracts the hyperparameters of the model.

        Returns
        ----------
        dict:
           The hyperparameter of the model.
        """
        pass

    def info(self):
        """Extracts the taxonomy metadata of the model.

        Returns
        ----------
        dict:
           The taxonomy metadata of the model.
        """
        return {
            MetadataTaxonomyKeys.FRAMEWORK: self.framework,
            MetadataTaxonomyKeys.FRAMEWORK_VERSION: self.version,
            MetadataTaxonomyKeys.ALGORITHM: str(self.algorithm),
            MetadataTaxonomyKeys.HYPERPARAMETERS: self.hyperparameter,
        }



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import datetime
import logging
import os
import fsspec
import oci
from typing import Any, Dict, List, Optional
from ads.common.oci_datascience import OCIDataScienceMixin
from ads.common.oci_logging import OCILog
from ads.common.oci_resource import ResourceNotFoundError
from ads.common import utils
from ads.jobs.builders.base import Builder
from ads.jobs.builders.infrastructure.dsc_job import DSCJob, DataScienceJob
from ads.jobs.builders.infrastructure.dsc_job_runtime import (
    DataScienceJobRuntimeManager,
)
from ads.jobs.builders.runtimes.artifact import Artifact
from ads.pipeline.ads_pipeline_step import PipelineStep
from ads.pipeline.ads_pipeline_run import PipelineRun
from ads.pipeline.visualizer.base import PipelineVisualizer, GraphOrientation
from ads.pipeline.visualizer.graph_renderer import PipelineGraphRenderer
from ads.jobs.builders.runtimes.python_runtime import (
    CondaRuntime,
    ScriptRuntime,
    PythonRuntime,
    NotebookRuntime,
    GitPythonRuntime,
)

logger = logging.getLogger(__name__)

MAXIMUM_TIMEOUT = 1800
DEFAULT_WAITER_KWARGS = {"max_wait_seconds": MAXIMUM_TIMEOUT}
DEFAULT_OPERATION_KWARGS = {
    "delete_related_pipeline_runs": True,
    "delete_related_job_runs": True,
}
ALLOWED_OPERATION_KWARGS = [
    "allow_control_chars",
    "retry_strategy",
    "delete_related_job_runs",
    "delete_related_pipeline_runs",
    "if_match",
    "opc_request_id",
]
ALLOWED_WAITER_KWARGS = [
    "max_interval_seconds",
    "max_wait_seconds",
    "succeed_on_not_found",
    "wait_callback",
    "fetch_func",
]


class Pipeline(Builder):
    """Represents a Data Science Machine Learning Pipeline."""

    CONST_ID = "id"
    CONST_PIPELINE_ID = "pipelineId"
    CONST_DISPLAY_NAME = "displayName"
    CONST_STEP_DETAILS = "stepDetails"
    CONST_STEP_OVERRIDE_DETAILS = "stepOverrideDetails"
    CONST_COMPARTMENT_ID = "compartmentId"
    CONST_PROJECT_ID = "projectId"
    CONST_LOG_GROUP_ID = "logGroupId"
    CONST_LOG_ID = "logId"
    CONST_SERVICE_LOG_ID = "serviceLogId"
    CONST_ENABLE_SERVICE_LOG = "enableServiceLog"
    CONST_ENVIRONMENT_VARIABLES = "environmentVariables"
    CONST_COMMAND_LINE_ARGUMENTS = "commandLineArguments"
    CONST_CREATED_BY = "createdBy"
    CONST_DESCRIPTION = "description"
    CONST_TYPE = "type"
    CONST_MAXIMUM_RUNTIME_IN_MINUTES = "maximumRuntimeInMinutes"
    CONST_ENABLE_LOGGING = "enableLogging"
    CONST_ENABLE_AUTO_LOG_CREATION = "enableAutoLogCreation"
    CONST_CONFIGURATION_DETAILS = "configurationDetails"
    CONST_CONFIGURATION_OVERRIDE_DETAILS = "configurationOverrideDetails"
    CONST_LOG_CONFIGURATION_DETAILS = "logConfigurationDetails"
    CONST_LOG_CONFIGURATION_OVERRIDE_DETAILS = "logConfigurationOverrideDetails"
    CONST_FREEFROM_TAGS = "freeformTags"
    CONST_DEFINED_TAGS = "definedTags"
    CONST_SYSTEM_TAGS = "systemTags"
    CONST_INFRA_CONFIG_DETAILS = "infrastructureConfigurationDetails"
    CONST_SHAPE_NAME = "shapeName"
    CONST_BLOCK_STORAGE_SIZE = "blockStorageSizeInGBs"
    CONST_SHAPE_CONFIG_DETAILS = "shapeConfigDetails"
    CONST_OCPUS = "ocpus"
    CONST_MEMORY_IN_GBS = "memoryInGBs"
    CONST_SERVICE_LOG_CATEGORY = "pipelinerunlog"
    CONST_SERVICE = "datascience"
    CONST_DAG = "dag"

    LIFECYCLE_STATE_CREATING = "CREATING"
    LIFECYCLE_STATE_ACTIVE = "ACTIVE"
    LIFECYCLE_STATE_DELETING = "DELETING"
    LIFECYCLE_STATE_FAILED = "FAILED"
    LIFECYCLE_STATE_DELETED = "DELETED"

    def __init__(self, name: str = None, spec: Dict = None, **kwargs) -> None:
        """Initialize a pipeline.

        Parameters
        ----------
        name: str
            The name of the pipeline, default to None. If a name is not provided, a randomly generated easy to remember
            name with timestamp will be generated, like 'strange-spider-2022-08-17-23:55.02'.
        spec : dict, optional
            Object specification, default to None
        kwargs: dict
            Specification as keyword arguments.
            If spec contains the same key as the one in kwargs, the value from kwargs will be used.

            - project_id: str
            - compartment_id: str
            - display_name: str
            - description: str
            - maximum_runtime_in_minutes: int
            - environment_variables: dict(str, str)
            - command_line_arguments: str
            - log_id: str
            - log_group_id: str
            - enable_service_log: bool
            - shape_name: str
            - block_storage_size_in_gbs: int
            - shape_config_details: dict
            - step_details: list[PipelineStep]
            - dag: list[str]
            - defined_tags: dict(str, dict(str, object))
            - freeform_tags: dict[str, str]

        Attributes
        ----------
        kind: str
            The kind of the object as showing in YAML.
        name: str
            The name of pipeline.
        id: str
            The id of pipeline.
        step_details: List[PipelineStep]
            The step details of pipeline.
        dag_details: List[str]
            The dag details of pipeline.
        log_group_id: str
            The log group id of pipeline.
        log_id: str
            The log id of pipeline.
        project_id: str
            The project id of pipeline.
        compartment_id: str
            The compartment id of pipeline.
        created_by: str
            The created by of pipeline.
        description: str
            The description of pipeline.
        environment_variable: dict
            The environment variables of pipeline.
        argument: str
            The command line argument of pipeline.
        maximum_runtime_in_minutes: int
            The maximum runtime in minutes of pipeline.
        shape_name: str
            The shape name of pipeline infrastructure.
        block_storage_size_in_gbs: int
            The block storage of pipeline infrastructure.
        shape_config_details: dict
            The shape config details of pipeline infrastructure.
        enable_service_log: bool
            The value to enable service log or not.
        service_log_id: str
            The service log id of pipeline.
        status: str
            The status of the pipeline.

        Methods
        -------
        with_name(self, name: str) -> Pipeline
            Sets the name of pipeline.
        with_id(self, id: str) -> Pipeline
            Sets the ocid of pipeline.
        with_step_details(self, step_details: List[PipelineStep]) -> Pipeline
            Sets the step details of pipeline.
        with_dag_details(self, dag_details: List[str]) -> Pipeline
            Sets the dag details of pipeline.
        with_log_group_id(self, log_group_id: str) -> Pipeline
            Sets the log group id of pipeline.
        with_log_id(self, log_id: str) -> Pipeline
            Sets the log id of pipeline.
        with_project_id(self, project_id: str) -> Pipeline
            Sets the project id of pipeline.
        with_compartment_id(self, compartment_id: str) -> Pipeline
            Sets the compartment id of pipeline.
        with_created_by(self, created_by: str) -> Pipeline
            Sets the created by of pipeline.
        with_description(self, description: str) -> Pipeline
            Sets the description of pipeline.
        with_environment_variable(self, **kwargs) -> Pipeline
            Sets the environment variables of pipeline.
        with_argument(self, *args, **kwargs) -> Pipeline
            Sets the command line arguments of pipeline.
        with_maximum_runtime_in_minutes(self, maximum_runtime_in_minutes: int) -> Pipeline
            Sets the maximum runtime in minutes of pipeline.
        with_freeform_tags(self, freeform_tags: Dict) -> Pipeline
            Sets the freeform tags of pipeline.
        with_defined_tags(self, defined_tags: Dict) -> Pipeline
            Sets the defined tags of pipeline.
        with_shape_name(self, shape_name: str) -> Pipeline
            Sets the shape name of pipeline infrastructure.
        with_block_storage_size_in_gbs(self, block_storage_size_in_gbs: int) -> Pipeline
            Sets the block storage size of pipeline infrastructure.
        with_shape_config_details(self, shape_config_details: Dict) -> Pipeline
            Sets the shape config details of pipeline infrastructure.
        with_enable_service_log(self, enable_service_log: bool) -> Pipeline
            Sets the value to enable the service log of pipeline.
        to_dict(self) -> dict:
            Serializes the pipeline specifications to a dictionary.
        from_dict(cls, obj_dict: dict):
            Initializes the object from a dictionary.
        create(self, delete_if_fail: bool = True) -> Pipeline
            Creates an ADS pipeline.
        show(self, rankdir: str = GraphOrientation.TOP_BOTTOM)
            Render pipeline with step information in a graph.
        to_svg(self, uri: str = None, rankdir: str = GraphOrientation.TOP_BOTTOM, **kwargs) -> str:
            Renders pipeline as graph into SVG.
        run(self, display_name: Optional[str] = None, project_id: Optional[str] = None, compartment_id: Optional[str] = None, configuration_override_details: Optional[dict] = None, log_configuration_override_details: Optional[dict] = None, step_override_details: Optional[list] = None, free_form_tags: Optional[dict] = None, defined_tags: Optional[dict] = None, system_tags: Optional[dict] = None) -> PipelineRun
            Creates and/or overrides an ADS pipeline run.
        delete(self, delete_related_pipeline_runs: Optional[bool] = True, delete_related_job_runs: Optional[bool] = True, max_wait_seconds: Optional[int] = MAXIMUM_TIMEOUT, **kwargs) -> Pipeline
            Deletes an ADS pipeline run.
        from_ocid(cls, ocid: str) -> Pipeline
            Creates an ADS pipeline from ocid.
        from_id(cls, id: str) -> Pipeline
            Creates an ADS pipeline from ocid.
        to_yaml(self, uri=None, **kwargs)
            Returns Pipeline serialized as a YAML string
        from_yaml(cls, yaml_string=None, uri=None, **kwargs)
            Creates an Pipeline from YAML string provided or from URI location containing YAML string
        list(cls, compartment_id: Optional[str] = None, **kwargs) -> List[Pipeline]
            List pipelines in a given compartment.
        run_list(self, **kwargs) -> List[PipelineRun]
            Gets a list of runs of the pipeline.

        Example
        -------
        Here is an example for creating and running a pipeline using builder:

        .. code-block:: python

            from ads.pipeline import Pipeline, CustomScriptStep, ScriptRuntime
            # Define an OCI Data Science pipeline
            pipeline = (
                Pipeline(name="<pipeline_name>")
                .with_compartment_id("<compartment_id>")
                .with_project_id("<project_id>")
                .with_log_group_id("<log_group_id>")
                .with_log_id("<log_id>")
                .with_description("<description>")
                .with_maximum_runtime_in_minutes(200)
                .with_argument("argument", key="value")
                .with_environment_variable(env="value")
                .with_freeform_tags({"key": "value"})
                .with_step_details([
                    (
                        PipelineStep(name="PipelineStepOne")
                        .with_job_id("<job_id>")
                        .with_description("<description>")
                    ),
                    (
                        PipelineStep(name="PipelineStepTwo")
                        .with_infrastructure(
                            CustomScriptStep()
                            .with_shape_name("VM.Standard2.1")
                            .with_block_storage_size(50)
                        )
                        .with_runtime(
                            ScriptRuntime()
                            .with_source("oci://bucket_name@namespace/path/to/script.py")
                            .with_service_conda("tensorflow26_p37_cpu_v2")
                            .with_environment_variable(ENV="value")
                            .with_argument("argument", key="value")
                            .with_maximum_runtime_in_minutes(200)
                        )
                    )
                ])
                .with_dag_details(["PipelineStepOne >> PipelineStepTwo"])
            )
            # Create and Run the pipeline
            run = pipeline.create().run()
            # Stream the pipeline run outputs
            run.watch()

        See Also
        --------
        https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/user_guide/pipeline/index.html
        """
        self.attribute_set = {
            "id",
            "createdBy",
            "projectId",
            "compartmentId",
            "description",
            "pipelineId",
            "displayName",
            "configurationDetails",
            "logConfigurationDetails",
            "infrastructureConfigurationDetails",
            "stepDetails",
            "lifecycleState",
            "freeformTags",
            "definedTags",
            "systemTags",
            "dag",
            "logId",
            "logGroupId",
        }

        self.attribute_map = {
            "projectId": "project_id",
            "compartmentId": "compartment_id",
            "displayName": "display_name",
            "description": "description",
            "maximumRuntimeInMinutes": "maximum_runtime_in_minutes",
            "environmentVariables": "environment_variables",
            "commandLineArguments": "command_line_arguments",
            "logId": "log_id",
            "logGroupId": "log_group_id",
            "enableServiceLog": "enable_service_log",
            "stepDetails": "step_details",
            "freeformTags": "freeform_tags",
            "definedTags": "defined_tags",
            "shapeName": "shape_name",
            "blockStorageSizeInGBs": "block_storage_size_in_gbs",
            "shapeConfigDetails": "shape_config_details",
            "dag": "dag",
            "id": "id",
        }

        self._artifact_content_map = {}
        super().__init__(spec=spec, **kwargs)
        name = (
            name
            or self.get_spec(self.CONST_DISPLAY_NAME)
            or utils.get_random_name_for_resource()
        )
        self.set_spec(self.CONST_DISPLAY_NAME, name)
        if "dag" in kwargs:
            self.with_dag(kwargs.get("dag"))
        elif spec and "dag" in spec:
            self.with_dag(spec.get("dag"))
        self.data_science_pipeline = None
        self.service_logging = None

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            pipeline
        """
        return "pipeline"

    @property
    def name(self) -> str:
        """The name of the pipeline.

        Returns
        -------
        str
            The name of the pipeline.
        """
        return self.get_spec(self.CONST_DISPLAY_NAME)

    def with_name(self, name: str) -> "Pipeline":
        """Sets the name of pipeline.

        Parameters
        ----------
        name: str
            The name of pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_DISPLAY_NAME, name)

    @property
    def id(self) -> str:
        """The id of the pipeline.

        Returns
        -------
        str
            The id of the pipeline.
        """
        return self.get_spec(self.CONST_ID)

    def with_id(self, id: str) -> "Pipeline":
        """Sets the id of pipeline.

        Parameters
        ----------
        id: str
            The id of pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_ID, id)

    @property
    def step_details(self) -> List["PipelineStep"]:
        """The step details of the pipeline.

        Returns
        -------
        list
            The step details of the pipeline.
        """
        return self.get_spec(self.CONST_STEP_DETAILS)

    def with_step_details(self, step_details: List["PipelineStep"]) -> "Pipeline":
        """Sets the pipeline step details for the pipeline.

        Parameters
        ----------
        step_details: list
            A list of steps in the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        self.set_spec(self.CONST_DAG, None)
        return self.set_spec(self.CONST_STEP_DETAILS, step_details)

    @property
    def dag(self) -> List[str]:
        """The dag details of the pipeline.

        Returns
        -------
        list
            The dag details of the pipeline.
        """
        return self.get_spec(self.CONST_DAG)

    def with_dag(self, dag: List[str]) -> "Pipeline":
        """Sets the pipeline dag details for the pipeline.

        Parameters
        ----------
        dag: list
            A list of dag representing step dependencies in the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        self.set_spec(self.CONST_DAG, dag)

        if not self.step_details:
            raise ValueError("Pipeline step details must be specified first.")

        stepname_to_step_map = {x.name: x for x in self.step_details}
        updated_step_details = Pipeline._add_dag_to_node(dag, stepname_to_step_map)

        return self.set_spec(self.CONST_STEP_DETAILS, updated_step_details)

    @staticmethod
    def _add_dag_to_node(
        dag: List[str], stepname_to_step_map: dict
    ) -> List["PipelineStep"]:
        """Add dependencies to pipeline steps.

        Parameters
        ----------
        dag: list
            A list of dag representing step dependencies in the pipeline.
        stepname_to_step_map: dict
            A dict mapping PipelineStep name to the PipelineStep object.

        Returns
        -------
        List
            A list of PipelineStep.
        """

        dag_mapping = {x: [] for x in stepname_to_step_map}
        updated_step_details = []
        if dag:
            for dag_line in dag:
                dependency = [
                    x.strip().strip("()").split(",") for x in dag_line.split(">>")
                ]
                for i in range(len(dependency) - 1):
                    for node1 in dependency[i]:
                        for node2 in dependency[i + 1]:
                            node1 = node1.strip()
                            node2 = node2.strip()
                            if node1 not in stepname_to_step_map:
                                raise ValueError(
                                    f"Pipeline step with name {node1} does not exist. Please provide a valid step name."
                                )
                            if node2 not in stepname_to_step_map:
                                raise ValueError(
                                    f"Pipeline step with name {node2} does not exist. Please provide a valid step name."
                                )

                            dag_mapping[node2].append(node1)

        for node, prev_list in dag_mapping.items():
            node_list = [stepname_to_step_map[x] for x in prev_list]
            stepname_to_step_map[node]._with_depends_on(node_list)
            updated_step_details.append(stepname_to_step_map[node])

        return updated_step_details

    @property
    def log_group_id(self) -> str:
        """The log group id of the pipeline.

        Returns
        -------
        str:
            The log group id of the pipeline.
        """
        return self.get_spec(self.CONST_LOG_GROUP_ID)

    def with_log_group_id(self, log_group_id: str) -> "Pipeline":
        """Sets the log group id of the pipeline.

        Parameters
        ----------
        log_group_id: str
            The log group id of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_LOG_GROUP_ID, log_group_id)

    @property
    def log_id(self) -> str:
        """The log id of the pipeline.

        Returns
        -------
        str:
            The log id of the pipeline.
        """
        return self.get_spec(self.CONST_LOG_ID)

    def with_log_id(self, log_id: str) -> "Pipeline":
        """Sets the log id of the pipeline.

        Parameters
        ----------
        log_id: str
            The log id of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_LOG_ID, log_id)

    @property
    def project_id(self) -> str:
        """The project id of the pipeline.

        Returns
        -------
        str:
            The project id of the pipeline.
        """
        return self.get_spec(self.CONST_PROJECT_ID)

    def with_project_id(self, project_id: str) -> "Pipeline":
        """Sets the project id of the pipeline.

        Parameters
        ----------
        project_id: str
            The project id of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_PROJECT_ID, project_id)

    @property
    def compartment_id(self) -> str:
        """The compartment id of the pipeline.

        Returns
        -------
        str:
            The compartment id of the pipeline.
        """
        return self.get_spec(self.CONST_COMPARTMENT_ID)

    def with_compartment_id(self, compartment_id: str) -> "Pipeline":
        """Sets the compartment id of the pipeline.

        Parameters
        ----------
        compartment_id: str
            The compartment id of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, compartment_id)

    @property
    def created_by(self) -> str:
        """The id that creates the pipeline.

        Returns
        -------
        str:
            The id that creates the pipeline.
        """
        return self.get_spec(self.CONST_CREATED_BY)

    def with_created_by(self, created_by: str) -> "Pipeline":
        """Sets the id that creates the pipeline.

        Parameters
        ----------
        created_by: str
            The id that creates the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_CREATED_BY, created_by)

    @property
    def description(self) -> str:
        """The description of pipeline.

        Returns
        -------
        str:
            The description of pipeline.
        """
        return self.get_spec(self.CONST_DESCRIPTION)

    def with_description(self, description: str) -> "Pipeline":
        """Sets the description of the pipeline.

        Parameters
        ----------
        description: str
            The description of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_DESCRIPTION, description)

    @property
    def environment_variable(self) -> dict:
        """The environment variables of the pipeline.

        Returns
        -------
        dict:
            The environment variables of the pipeline.
        """
        return self.get_spec(self.CONST_ENVIRONMENT_VARIABLES)

    def with_environment_variable(self, **kwargs) -> "Pipeline":
        """Sets environment variables of the pipeline.

        Parameters
        ----------
        kwargs:
            Keyword arguments.
            To add a keyword argument without value, set the value to None.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        if kwargs:
            environment_variable_dict = {}
            for k, v in kwargs.items():
                environment_variable_dict[k] = v
            self.set_spec(self.CONST_ENVIRONMENT_VARIABLES, environment_variable_dict)
        return self

    @property
    def argument(self) -> str:
        """The command line arguments of the pipeline.

        Returns
        -------
        str:
            The command line arguments of the pipeline.
        """
        return self.get_spec(self.CONST_COMMAND_LINE_ARGUMENTS)

    def with_argument(self, *args, **kwargs) -> "Pipeline":
        """Adds command line arguments to the pipeline.
        Existing arguments will be preserved.
        This method can be called (chained) multiple times to add various arguments.
        For example, pipeline.with_argument(key="val").with_argument("path/to/file")
        will result in: "--key val path/to/file"

        Parameters
        ----------
        args:
            Positional arguments.
            In a single method call, positional arguments are always added before keyword arguments.
            You can call with_argument() to add positional arguments after keyword arguments.

        kwargs:
            Keyword arguments.
            To add a keyword argument without value, set the value to None.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).

        Raises
        ------
        ValueError
            Keyword arguments with space in a key.
        """
        arg_values = self.get_spec(self.CONST_COMMAND_LINE_ARGUMENTS, [])
        args = [str(arg) for arg in args]
        arg_values.extend(args)
        for k, v in kwargs.items():
            if " " in k:
                raise ValueError("Argument key %s cannot contain space.", str(k))
            arg_values.append(f"--{str(k)}")
            # Ignore None value
            if v is None:
                continue
            arg_values.append(str(v))
        arg_string = " ".join(arg_values)
        self.set_spec(self.CONST_COMMAND_LINE_ARGUMENTS, arg_string)
        return self

    @property
    def maximum_runtime_in_minutes(self) -> int:
        """The maximum runtime in minutes of the pipeline.

        Returns
        -------
        int:
            The maximum runtime minutes of the pipeline.
        """
        return self.get_spec(self.CONST_MAXIMUM_RUNTIME_IN_MINUTES)

    def with_maximum_runtime_in_minutes(
        self, maximum_runtime_in_minutes: int
    ) -> "Pipeline":
        """Sets the maximum runtime in minutes of the pipeline.

        Parameters
        ----------
        maximum_runtime_in_minutes: int
            The maximum_runtime_in_minutes of the pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(
            self.CONST_MAXIMUM_RUNTIME_IN_MINUTES, maximum_runtime_in_minutes
        )

    def with_freeform_tags(self, freeform_tags: Dict) -> "Pipeline":
        """Sets freeform tags of the pipeline.

        Parameters
        ----------
        freeform_tags: dict
            The freeform tags dictionary.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_FREEFROM_TAGS, freeform_tags)

    def with_defined_tags(self, defined_tags: Dict) -> "Pipeline":
        """Sets defined tags of the pipeline.

        Parameters
        ----------
        defined_tags: dict
            The defined tags dictionary.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_DEFINED_TAGS, defined_tags)

    @property
    def shape_name(self) -> str:
        """The shape name of pipeline infrastructure.

        Returns
        -------
        str:
            The shape name of the pipeline infrastructure.
        """
        return self.get_spec(self.CONST_SHAPE_NAME)

    def with_shape_name(self, shape_name: str) -> "Pipeline":
        """Sets the shape name of pipeline infrastructure.

        Parameters
        ----------
        shape_name: str
            The shape name of the pipeline infrastructure.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_SHAPE_NAME, shape_name)

    @property
    def block_storage_size_in_gbs(self) -> int:
        """The block storage size of pipeline infrastructure.

        Returns
        -------
        int:
            The block storage size of the pipeline infrastructure.
        """
        return self.get_spec(self.CONST_BLOCK_STORAGE_SIZE)

    def with_block_storage_size_in_gbs(
        self, block_storage_size_in_gbs: int
    ) -> "Pipeline":
        """Sets the block storage size of pipeline infrastructure.

        Parameters
        ----------
        block_storage_size_in_gbs: int
            The block storage size of pipeline infrastructure.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_BLOCK_STORAGE_SIZE, block_storage_size_in_gbs)

    @property
    def shape_config_details(self) -> dict:
        """The shape config details of pipeline infrastructure.

        Returns
        -------
        dict:
            The shape config details of the pipeline infrastructure.
        """
        return self.get_spec(self.CONST_SHAPE_CONFIG_DETAILS)

    def with_shape_config_details(
        self, memory_in_gbs: float, ocpus: float, **kwargs: Dict[str, Any]
    ) -> "Pipeline":
        """
        Sets the shape config details of pipeline infrastructure.
        Specify only when a flex shape is selected.
        For example `VM.Standard.E3.Flex` allows the memory_in_gbs and cpu count to be specified.

        Parameters
        ----------
        memory_in_gbs: float
            The size of the memory in GBs.
        ocpus: float
            The OCPUs count.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(
            self.CONST_SHAPE_CONFIG_DETAILS,
            {
                self.CONST_OCPUS: ocpus,
                self.CONST_MEMORY_IN_GBS: memory_in_gbs,
                **kwargs,
            },
        )

    @property
    def enable_service_log(self) -> bool:
        """Enables service log of pipeline.

        Returns
        -------
        bool:
            The bool value to enable service log of pipeline.
        """
        return self.get_spec(self.CONST_ENABLE_SERVICE_LOG)

    def with_enable_service_log(self, enable_service_log: bool) -> "Pipeline":
        """Sets the bool value to enable the service log of pipeline.

        Parameters
        ----------
        enable_service_log: bool
            The value to enable the service log of pipeline.

        Returns
        -------
        Pipeline
            The Pipeline instance (self).
        """
        return self.set_spec(self.CONST_ENABLE_SERVICE_LOG, enable_service_log)

    @property
    def service_log_id(self) -> str:
        """The service log id of pipeline.

        Returns
        -------
        str:
            The service log id of pipeline.
        """
        return self.get_spec(self.CONST_SERVICE_LOG_ID)

    def to_dict(self) -> dict:
        """Serializes the pipeline specifications to a dictionary.

        Returns
        -------
        dict
            A dictionary containing pipeline specifications.
        """
        dict_details = copy.deepcopy(super().to_dict())
        dict_details["spec"][self.CONST_DAG] = self.get_spec(self.CONST_DAG)

        step_details_list = []
        if self.step_details:
            for step in self.step_details:
                if isinstance(step, PipelineStep):
                    step = step.to_dict()

                if not isinstance(step, dict):
                    raise TypeError("Pipeline step is not a valid type")
                step_dict = copy.deepcopy(step)
                step_dict["spec"].pop("dependsOn", None)
                step_details_list.append(step_dict)

        dict_details["spec"][self.CONST_STEP_DETAILS] = step_details_list
        return dict_details

    @classmethod
    def from_dict(cls, obj_dict: dict):
        """Initializes the object from a dictionary."""
        temp_obj_dict = copy.deepcopy(obj_dict)
        step_mapping = {}
        for step in temp_obj_dict["spec"][cls.CONST_STEP_DETAILS]:
            pipeline_step = PipelineStep.from_dict(step)
            step_mapping[pipeline_step.name] = pipeline_step

        if cls.CONST_DAG not in temp_obj_dict["spec"]:
            temp_obj_dict["spec"][cls.CONST_DAG] = None

        step_details = Pipeline._add_dag_to_node(
            temp_obj_dict["spec"][cls.CONST_DAG], step_mapping
        )
        temp_obj_dict["spec"][cls.CONST_STEP_DETAILS] = step_details

        return cls(spec=temp_obj_dict["spec"])

    def create(self, delete_if_fail: bool = True) -> "Pipeline":
        """Creates an ADS pipeline.

        Returns
        -------
        Pipeline:
            The ADS Pipeline instance.
        """
        pipeline_details = self.__pipeline_details()
        self.data_science_pipeline = DataSciencePipeline(**pipeline_details).create(
            self.step_details, delete_if_fail
        )
        self.set_spec(self.CONST_ID, self.data_science_pipeline.id)
        if self.enable_service_log and not self.service_log_id:
            try:
                self.__create_service_log()
            except Exception as ex:
                logger.warning("Failed to create service log: %s", str(ex))
        return self

    def _show(self) -> PipelineVisualizer:
        """
        Prepeares `PipelineVisualizer` instance to render a graph.

        Returns
        -------
        PipelineVisualizer
        """
        return (
            PipelineVisualizer()
            .with_renderer(PipelineGraphRenderer(show_status=False))
            .with_pipeline(self)
        )

    def show(self, rankdir: str = GraphOrientation.TOP_BOTTOM) -> None:
        """
        Render pipeline with step information in a graph

        Returns
        -------
        None
        """
        self._show().render(rankdir=rankdir)

    def to_svg(
        self, uri: str = None, rankdir: str = GraphOrientation.TOP_BOTTOM, **kwargs
    ) -> str:
        """
        Renders pipeline as graph in svg string.

        Parameters
        ----------
        uri: (string, optional). Defaults to None.
            URI location to save the SVG string.
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.

        Returns
        -------
        str
            Graph in svg format.
        """
        return self._show().to_svg(uri=uri, rankdir=rankdir, **kwargs)

    def run(
        self,
        display_name: Optional[str] = None,
        project_id: Optional[str] = None,
        compartment_id: Optional[str] = None,
        configuration_override_details: Optional[dict] = None,
        log_configuration_override_details: Optional[dict] = None,
        step_override_details: Optional[list] = None,
        free_form_tags: Optional[dict] = None,
        defined_tags: Optional[dict] = None,
        system_tags: Optional[dict] = None,
    ) -> "PipelineRun":
        """Creates an ADS pipeline run.

        Parameters
        ----------
        display_name: str, optional
            The display name to override the one defined previously. Defaults to None.

        project_id: str, optional
            The project id to override the one defined previously. Defaults to None.

        compartment_id: str, optional
            The compartment id to override the one defined previously. Defaults to None.

        configuration_override_details: dict, optional
            The configuration details dictionary to override the one defined previously.
            Defaults to None.
            The configuration_override_details contains the following keys:
            * "type": str, only "DEFAULT" is allowed.
            * "environment_variables": dict, optional, the environment variables
            * "command_line_arguments": str, optional, the command line arguments
            * "maximum_runtime_in_minutes": int, optional, the maximum runtime allowed in minutes

        log_configuration_override_details: dict(str, str), optional
            The log configuration details dictionary to override the one defined previously.
            Defaults to None.
            The log_configuration_override_details contains the following keys:
            * "log_group_id": str, optional, the log group id
            * "log_id": str, optional, the log id

        step_override_details: list[PipelineStepOverrideDetails], optional
            The step details list to override the one defined previously.
            Defaults to None.
            The PipelineStepOverrideDetails is a dict which contains the following keys:
            * step_name: str, the name of step to override
            * step_configuration_details: dict, which contains:
                * "maximum_runtime_in_minutes": int, optional
                * "environment_variables": dict, optional
                * "command_line_arguments": str, optional

        free_form_tags: dict(str, str), optional
            The free from tags dictionary to override the one defined previously.
            Defaults to None.

        defined_tags: dict(str, dict(str, object)), optional
            The defined tags dictionary to override the one defined previously.
            Defaults to None.

        system_tags: dict(str, dict(str, object)), optional
            The system tags dictionary to override the one defined previously.
            Defaults to None.

        Example
        --------
        .. code-block:: python

            # Creates a pipeline run using pipeline configurations
            pipeline.run()

            # Creates a pipeline run by overriding pipeline configurations
            pipeline.run(
                display_name="OverrideDisplayName",
                configuration_override_details={
                    "maximum_runtime_in_minutes":30,
                    "type":"DEFAULT",
                    "environment_variables": {
                        "key": "value"
                    },
                    "command_line_arguments": "ARGUMENT --KEY VALUE",
                },
                log_configuration_override_details={
                    "log_group_id": "<log_group_id>"
                },
                step_override_details=[{
                    "step_name" : "<step_name>",
                    "step_configuration_details" : {
                        "maximum_runtime_in_minutes": 200,
                        "environment_variables": {
                            "1":"2"
                        },
                        "command_line_arguments": "argument --key value",
                    }
                }]
            )

        Returns
        -------
        PipelineRun:
            The ADS PipelineRun instance.
        """
        pipeline_details = self.__pipeline_details()
        self.__override_configurations(
            pipeline_details,
            display_name,
            project_id,
            compartment_id,
            configuration_override_details,
            log_configuration_override_details,
            step_override_details,
            free_form_tags,
            defined_tags,
            system_tags,
        )
        if not self.data_science_pipeline:
            self.data_science_pipeline = DataSciencePipeline(**pipeline_details)

        if self.enable_service_log:
            return self.data_science_pipeline.run(
                pipeline_details, self.service_logging
            )

        return self.data_science_pipeline.run(pipeline_details)

    def delete(
        self,
        delete_related_pipeline_runs: Optional[bool] = True,
        delete_related_job_runs: Optional[bool] = True,
        max_wait_seconds: Optional[int] = MAXIMUM_TIMEOUT,
        **kwargs,
    ) -> "Pipeline":
        """Deteles an ADS pipeline.

        Parameters
        ----------
        delete_related_pipeline_runs: bool, optional
            Specify whether to delete related PipelineRuns or not. Defaults to True.
        delete_related_job_runs: bool, optional
            Specify whether to delete related JobRuns or not. Defaults to True.
        max_wait_seconds: int, optional
            The maximum time to wait, in seconds. Defaults to 1800.

        kwargs: optional
        The kwargs to be executed when deleting the pipeline.
        The allowed keys are:
        * "allow_control_chars": bool, to indicate whether or not this request should
        allow control characters in the response object. By default, the response will
        not allow control characters in strings.
        * "retry_strategy": obj, to apply to this specific operation/call. This will
        override any retry strategy set at the client-level. This should be one of the
        strategies available in the :py:mod:`~oci.retry` module. This operation will not
        retry by default, users can also use the convenient :py:data:`~oci.retry.DEFAULT_RETRY_STRATEGY`
        provided by the SDK to enable retries for it. The specifics of the default retry
        strategy are described `here <https://docs.oracle.com/en-us/iaas/tools/python/latest/sdk_behaviors/retries.html>`__.
        To have this operation explicitly not perform any retries, pass an instance of :py:class:`~oci.retry.NoneRetryStrategy`.
        * "if_match": str, for optimistic concurrency control. In the PUT or DELETE call
        for a resource, set the `if-match` parameter to the value of the etag from a
        previous GET or POST response for that resource. The resource is updated or
        deleted only if the `etag` you provide matches the resource's current `etag` value.
        * "opc_request_id": str, unique Oracle assigned identifier for the request.
        If you need to contact Oracle about a particular request, then provide the request ID.
        * "max_interval_seconds": int, the maximum interval between queries, in seconds.
        * "succeed_on_not_found": bool, to determine whether or not the waiter should
        return successfully if the data we're waiting on is not found
        (e.g. a 404 is returned from the service). This defaults to False and so a 404 would
        cause an exception to be thrown by this function. Setting it to True may be useful in
        scenarios when waiting for a resource to be terminated/deleted since it is possible that
        the resource would not be returned by the a GET call anymore.
        * "wait_callback": A function which will be called each time that we have to do an initial
        wait (i.e. because the property of the resource was not in the correct state,
        or the ``evaluate_response`` function returned False). This function should take two
        arguments - the first argument is the number of times we have checked the resource,
        and the second argument is the result of the most recent check.
        * "fetch_func": A function to be called to fetch the updated state from the server.
        This can be used if the call to check for state needs to be more complex than a single
        GET request. For example, if the goal is to wait until an item appears in a list,
        fetch_func can be a function that paginates through a full list on the server.

        Returns
        -------
        Pipeline:
            The ADS Pipeline instance.
        """
        if not self.data_science_pipeline:
            self.data_science_pipeline = DataSciencePipeline()

        operation_kwargs = {
            "delete_related_pipeline_runs": delete_related_pipeline_runs,
            "delete_related_job_runs": delete_related_job_runs,
        }
        waiter_kwargs = {"max_wait_seconds": max_wait_seconds}
        for key, value in kwargs.items():
            if key in ALLOWED_OPERATION_KWARGS:
                operation_kwargs[key] = value
            elif key in ALLOWED_WAITER_KWARGS:
                waiter_kwargs[key] = value

        self.data_science_pipeline.delete(
            id=self.id, operation_kwargs=operation_kwargs, waiter_kwargs=waiter_kwargs
        )
        return self

    def download(
        self, to_dir: str, override_if_exists: Optional[bool] = False
    ) -> "Pipeline":
        """Downloads artifacts from pipeline.

        Parameters
        ----------
        to_dir : str
            Local directory to which the artifacts will be downloaded to.
        override_if_exists: bool, optional
            Bool to decide whether to override existing folder/file or not. Defaults to False.

        Returns
        -------
        Pipeline:
            The ADS Pipeline instance.
        """
        if not self.data_science_pipeline or not self.id:
            print("Pipeline hasn't been created.")
            return self

        if (
            self.data_science_pipeline.sync().lifecycle_state
            != self.LIFECYCLE_STATE_ACTIVE
        ):
            print("Pipeline hasn't been created or not in ACTIVE state.")
            return self

        pipeline_folder = os.path.join(to_dir, self.id)
        if not os.path.exists(pipeline_folder):
            print("Creating directory: " + pipeline_folder)
            os.mkdir(pipeline_folder)
        elif not override_if_exists:
            print(
                f"Folder {pipeline_folder} already exists. Set override_if_exists to True to override."
            )
            return self

        for step in self.step_details:
            if step.kind == "CUSTOM_SCRIPT":
                res = self.data_science_pipeline.client.get_step_artifact_content(
                    self.id, step.name
                )

                if not res or not res.data or not res.data.raw:
                    print(f"Failed to download {step.name} artifact.")
                    return self

                content_disposition = res.headers.get("Content-Disposition", "")
                artifact_name = str(content_disposition).replace(
                    "attachment; filename=", ""
                )
                step_folder = os.path.join(pipeline_folder, step.name)
                if not os.path.exists(step_folder):
                    print("Creating directory: " + step_folder)
                    os.mkdir(step_folder)
                elif not override_if_exists:
                    print(
                        f"Folder {step_folder} already exists. Set override_if_exists to True to override."
                    )
                    continue
                file_name = os.path.join(step_folder, artifact_name)
                with open(file_name, "wb") as f:
                    f.write(res.data.raw.read())

        return self

    def _populate_step_artifact_content(self):
        """Populates artifact information to CUSTOM_SCRIPT step.
        This method is only invoked when the existing pipeline needs to be loaded.
        """
        if (
            not self.data_science_pipeline
            or not self.id
            or self.status != self.LIFECYCLE_STATE_ACTIVE
        ):
            return

        for step in self.step_details:
            if step.kind == "CUSTOM_SCRIPT":
                artifact_name = self._artifact_content_map.get(step.name)
                if not artifact_name:
                    res = self.data_science_pipeline.client.get_step_artifact_content(
                        self.id, step.name
                    )
                    content_disposition = res.headers.get("Content-Disposition", "")
                    artifact_name = str(content_disposition).replace(
                        "attachment; filename=", ""
                    )
                    self._artifact_content_map[step.name] = artifact_name
                if isinstance(step.runtime, ScriptRuntime):
                    step.runtime.with_script(artifact_name)
                elif isinstance(step.runtime, PythonRuntime):
                    step.runtime.with_working_dir(artifact_name)
                elif isinstance(step.runtime, NotebookRuntime):
                    step.runtime.with_notebook(artifact_name)
                elif isinstance(step.runtime, GitPythonRuntime):
                    step.runtime.with_source(artifact_name)

    @classmethod
    def from_ocid(cls, ocid: str) -> "Pipeline":
        """Creates a pipeline by OCID.

        Parameters
        ----------
        ocid: str
            The OCID of pipeline.

        Returns
        -------
        Pipeline:
            The Pipeline instance.
        """
        pipeline = DataSciencePipeline.from_ocid(ocid).build_ads_pipeline()
        pipeline._populate_step_artifact_content()
        return pipeline

    @classmethod
    def from_id(cls, id: str) -> "Pipeline":
        """Creates a pipeline by OCID.

        Parameters
        ----------
        id: str
            The OCID of pipeline.

        Returns
        -------
        Pipeline:
            The Pipeline instance.
        """
        return cls.from_ocid(id)

    def __create_service_log(self) -> "Pipeline":
        """Creates a service log for pipeline.

        Returns
        -------
        Pipeline:
            The ADS Pipeline instance.
        """
        if not self.log_group_id:
            raise ValueError(
                "Log group OCID is not specified for this pipeline. Call with_log_group_id to add it."
            )
        if not self.id:
            raise ValueError("Pipeline is not created yet. Call the create method.")

        oci_service = oci.logging.models.OciService(
            service=self.CONST_SERVICE,
            resource=self.id,
            category=self.CONST_SERVICE_LOG_CATEGORY,
        )

        archiving = oci.logging.models.Archiving(is_enabled=False)

        configuration = oci.logging.models.Configuration(
            source=oci_service, compartment_id=self.compartment_id, archiving=archiving
        )

        timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M")
        service_logging = OCILog(
            display_name=self.name + f"-{timestamp}",
            log_group_id=self.log_group_id,
            log_type="SERVICE",
            configuration=configuration,
            annotation="service",
        ).create()

        self.__set_service_logging_resource(service_logging)
        self.set_spec(self.CONST_SERVICE_LOG_ID, self.service_logging.id)
        return self

    def __set_service_logging_resource(self, service_logging: OCILog):
        """Sets the service logging for pipeline.

        Parameters
        ----------
        service_logging: OCILog
            An OCILog instance containing the service logging resources.
        """
        self.service_logging = service_logging

    def _convert_step_details_to_dag(
        self, step_details: List["PipelineStep"] = []
    ) -> List[str]:
        """Converts step_details to the DAG representation.

        Parameters
        ----------
        step_details: list
            A list of PipelineStep objects, default to empty list.

        Returns
        -------
        List
            A list of str representing step dependencies.
        """
        dag_list = []
        for step in step_details:
            if isinstance(step, PipelineStep):
                step = step.to_dict()["spec"]
                step_name = step["name"]
            else:
                step_name = step["stepName"]

            if not step["dependsOn"]:
                continue
            if len(step["dependsOn"]) == 1:
                dag = step["dependsOn"][0] + " >> " + step_name
            else:
                dag = "(" + ", ".join(step["dependsOn"]) + ") >> " + step_name
            dag_list.append(dag)

        return dag_list

    def __pipeline_details(self) -> dict:
        """Converts pipeline attributes to a dictionary.

        Returns
        -------
        dict:
            A dictionary that contains pipeline details.
        """
        pipeline_details = copy.deepcopy(self._spec)
        pipeline_details.pop(self.CONST_ENABLE_SERVICE_LOG, None)
        pipeline_details.pop(self.CONST_SERVICE_LOG_ID, None)
        pipeline_configuration_details = self.__pipeline_configuration_details(
            pipeline_details
        )
        if pipeline_configuration_details:
            pipeline_details[
                self.CONST_CONFIGURATION_DETAILS
            ] = pipeline_configuration_details

        pipeline_log_configuration_details = self.__pipeline_log_configuration_details(
            pipeline_details
        )
        if pipeline_log_configuration_details:
            pipeline_details[
                self.CONST_LOG_CONFIGURATION_DETAILS
            ] = pipeline_log_configuration_details

        pipeline_infrastructure_configuration_details = (
            self.__pipeline_infrastructure_configuration_details(pipeline_details)
        )
        if pipeline_infrastructure_configuration_details:
            pipeline_details[
                self.CONST_INFRA_CONFIG_DETAILS
            ] = pipeline_infrastructure_configuration_details

        if self.id:
            pipeline_details[self.CONST_PIPELINE_ID] = self.id
            pipeline_details.pop(self.CONST_ID)

        step_details_list = self.__step_details(pipeline_details)
        pipeline_details[self.CONST_STEP_DETAILS] = step_details_list

        return pipeline_details

    def __pipeline_configuration_details(self, pipeline_details: Dict) -> dict:
        """Converts pipeline configuration details to a dictionary.

        Parameters
        ----------
        pipeline_details: dict
            A dictionary that contains pipeline details.

        Returns
        -------
        dict:
            A dictionary that contains pipeline configuration details.
        """
        pipeline_configuration_details = {}
        if self.maximum_runtime_in_minutes:
            pipeline_configuration_details[
                self.CONST_MAXIMUM_RUNTIME_IN_MINUTES
            ] = self.maximum_runtime_in_minutes
            pipeline_details.pop(self.CONST_MAXIMUM_RUNTIME_IN_MINUTES)
        if self.environment_variable:
            pipeline_configuration_details[
                self.CONST_ENVIRONMENT_VARIABLES
            ] = self.environment_variable
            pipeline_details.pop(self.CONST_ENVIRONMENT_VARIABLES)
        if self.argument:
            pipeline_configuration_details[
                self.CONST_COMMAND_LINE_ARGUMENTS
            ] = self.argument
            pipeline_details.pop(self.CONST_COMMAND_LINE_ARGUMENTS)
        pipeline_configuration_details[self.CONST_TYPE] = "DEFAULT"
        return pipeline_configuration_details

    def __pipeline_log_configuration_details(self, pipeline_details: Dict) -> dict:
        """Converts pipeline log configuration details to a dictionary.

        Parameters
        ----------
        pipeline_details: dict
            A dictionary that contains pipeline details.

        Returns
        -------
        dict:
            A dictionary that contains pipeline log configuration details.
        """
        pipeline_log_configuration_details = {}
        if self.log_id:
            pipeline_log_configuration_details[self.CONST_LOG_ID] = self.log_id
            if not self.log_group_id:
                try:
                    log_obj = OCILog.from_ocid(self.log_id)
                except ResourceNotFoundError:
                    raise ResourceNotFoundError(
                        f"Unable to determine log group ID for Log ({self.log_id})."
                        " The log resource may not exist or You may not have the required permission."
                        " Try to avoid this by specifying the log group ID."
                    )
                self.with_log_group_id(log_obj.log_group_id)

        if self.log_group_id:
            pipeline_log_configuration_details[
                self.CONST_LOG_GROUP_ID
            ] = self.log_group_id

        if self.log_id:
            pipeline_log_configuration_details[self.CONST_ENABLE_LOGGING] = True
            pipeline_log_configuration_details[
                self.CONST_ENABLE_AUTO_LOG_CREATION
            ] = False
            pipeline_details.pop(self.CONST_LOG_ID)
            pipeline_details.pop(self.CONST_LOG_GROUP_ID, None)
        else:
            if self.log_group_id:
                pipeline_log_configuration_details[self.CONST_ENABLE_LOGGING] = True
                pipeline_log_configuration_details[
                    self.CONST_ENABLE_AUTO_LOG_CREATION
                ] = True
                pipeline_details.pop(self.CONST_LOG_GROUP_ID)
            else:
                pipeline_log_configuration_details[self.CONST_ENABLE_LOGGING] = False
                pipeline_log_configuration_details[
                    self.CONST_ENABLE_AUTO_LOG_CREATION
                ] = False
        return pipeline_log_configuration_details

    def __pipeline_infrastructure_configuration_details(
        self, pipeline_details: Dict
    ) -> dict:
        pipeline_infrastructure_details = {}
        if self.shape_name:
            pipeline_infrastructure_details[self.CONST_SHAPE_NAME] = self.shape_name
            pipeline_details.pop(self.CONST_SHAPE_NAME)
        if self.block_storage_size_in_gbs:
            pipeline_infrastructure_details[
                self.CONST_BLOCK_STORAGE_SIZE
            ] = self.block_storage_size_in_gbs
            pipeline_details.pop(self.CONST_BLOCK_STORAGE_SIZE)
        if self.shape_config_details:
            pipeline_infrastructure_details[
                self.CONST_SHAPE_CONFIG_DETAILS
            ] = self.shape_config_details
            pipeline_details.pop(self.CONST_SHAPE_CONFIG_DETAILS)

        return pipeline_infrastructure_details

    def __step_details(self, pipeline_details: Dict) -> list:
        """Converts pipeline step details to a dictionary.

        Parameters
        ----------
        pipeline_details: dict
            A dictionary that contains pipeline details.

        Returns
        -------
        list:
            A list that contains pipeline step details.
        """
        step_details_list = []
        if self.step_details:
            for step in self.step_details:
                step_details = copy.deepcopy(step._spec)
                step_details["stepName"] = step.name
                step_details.pop("name", None)
                if not step.depends_on:
                    step_details[step.CONST_DEPENDS_ON] = []
                if not step.job_id:
                    step_infrastructure_configuration_details = (
                        self.__step_infrastructure_configuration_details(step)
                    )
                    step_details[
                        step.CONST_STEP_INFRA_CONFIG_DETAILS
                    ] = step_infrastructure_configuration_details
                    step_details.pop(step.CONST_INFRASTRUCTURE, None)
                    step_details.pop(step.CONST_RUNTIME, None)

                step_configuration_details = self.__step_configuration_details(
                    pipeline_details, step
                )
                step_details[
                    step.CONST_STEP_CONFIG_DETAILS
                ] = step_configuration_details
                step_details.pop(self.CONST_MAXIMUM_RUNTIME_IN_MINUTES, None)
                step_details.pop(self.CONST_ENVIRONMENT_VARIABLES, None)
                step_details.pop(self.CONST_COMMAND_LINE_ARGUMENTS, None)
                step_details_list.append(step_details)
        return step_details_list

    def __step_infrastructure_configuration_details(self, step) -> dict:
        step_infrastructure_configuration_details = {}
        step_infrastructure_configuration_details[
            "blockStorageSizeInGBs"
        ] = step.infrastructure.block_storage_size
        step_infrastructure_configuration_details[
            "shapeName"
        ] = step.infrastructure.shape_name
        step_infrastructure_configuration_details[
            "shapeConfigDetails"
        ] = step.infrastructure.shape_config_details
        return step_infrastructure_configuration_details

    def __step_configuration_details(self, pipeline_details: Dict, step) -> dict:
        step_configuration_details = {}
        step_configuration_details[self.CONST_TYPE] = "DEFAULT"
        if step.runtime:
            payload = DataScienceJobRuntimeManager(step.infrastructure).translate(
                step.runtime
            )
            if "job_configuration_details" in payload:
                job_configuration_details = payload["job_configuration_details"]
                if "environment_variables" in job_configuration_details:
                    step_configuration_details[
                        self.CONST_ENVIRONMENT_VARIABLES
                    ] = job_configuration_details["environment_variables"]
                if "command_line_arguments" in job_configuration_details:
                    step_configuration_details[
                        self.CONST_COMMAND_LINE_ARGUMENTS
                    ] = job_configuration_details["command_line_arguments"]
                if "maximum_runtime_in_minutes" in job_configuration_details:
                    step_configuration_details[
                        self.CONST_MAXIMUM_RUNTIME_IN_MINUTES
                    ] = job_configuration_details["maximum_runtime_in_minutes"]
        elif step.CONST_STEP_CONFIG_DETAILS in step._spec:
            step_configuration_details = step._spec[step.CONST_STEP_CONFIG_DETAILS]

        if len(step_configuration_details) == 1:
            if step.environment_variable:
                step_configuration_details[
                    self.CONST_ENVIRONMENT_VARIABLES
                ] = step.environment_variable
            if step.argument:
                step_configuration_details[
                    self.CONST_COMMAND_LINE_ARGUMENTS
                ] = step.argument
            if step.maximum_runtime_in_minutes:
                step_configuration_details[
                    self.CONST_MAXIMUM_RUNTIME_IN_MINUTES
                ] = step.maximum_runtime_in_minutes

        if len(step_configuration_details) == 1:
            if self.CONST_CONFIGURATION_DETAILS in pipeline_details:
                step_configuration_details = pipeline_details[
                    self.CONST_CONFIGURATION_DETAILS
                ]

        return step_configuration_details

    def __override_configurations(
        self,
        pipeline_details,
        display_name,
        project_id,
        compartment_id,
        configuration_override_details,
        log_configuration_override_details,
        step_override_details,
        free_form_tags,
        defined_tags,
        system_tags,
    ) -> dict:
        if display_name:
            pipeline_details[self.CONST_DISPLAY_NAME] = display_name

        if project_id:
            pipeline_details[self.CONST_PROJECT_ID] = project_id

        if compartment_id:
            pipeline_details[self.CONST_COMPARTMENT_ID] = compartment_id

        if configuration_override_details:
            configuration_override_details[self.CONST_TYPE] = "DEFAULT"
            pipeline_details[
                self.CONST_CONFIGURATION_OVERRIDE_DETAILS
            ] = self._standardize_spec(configuration_override_details)

        if log_configuration_override_details:
            pipeline_details[
                self.CONST_LOG_CONFIGURATION_OVERRIDE_DETAILS
            ] = self._standardize_spec(log_configuration_override_details)
            log_configuration_override_details = pipeline_details[
                self.CONST_LOG_CONFIGURATION_OVERRIDE_DETAILS
            ]

            if (
                self.CONST_LOG_ID in log_configuration_override_details
                and self.CONST_LOG_GROUP_ID not in log_configuration_override_details
            ):
                try:
                    log_obj = OCILog.from_ocid(
                        log_configuration_override_details[self.CONST_LOG_ID]
                    )
                except ResourceNotFoundError:
                    raise ResourceNotFoundError(
                        f"Unable to determine log group ID for Log ({log_configuration_override_details[self.CONST_LOG_ID]})."
                        " The log resource may not exist or You may not have the required permission."
                        " Try to avoid this by specifying the log group ID."
                    )
                if log_obj and log_obj.log_group_id:
                    log_configuration_override_details[
                        self.CONST_LOG_GROUP_ID
                    ] = log_obj.log_group_id

            if self.CONST_LOG_ID in log_configuration_override_details:
                log_configuration_override_details[self.CONST_ENABLE_LOGGING] = True
                log_configuration_override_details[
                    self.CONST_ENABLE_AUTO_LOG_CREATION
                ] = False
            else:
                if self.CONST_LOG_GROUP_ID in log_configuration_override_details:
                    log_configuration_override_details[self.CONST_ENABLE_LOGGING] = True
                    log_configuration_override_details[
                        self.CONST_ENABLE_AUTO_LOG_CREATION
                    ] = True
                else:
                    log_configuration_override_details[
                        self.CONST_ENABLE_LOGGING
                    ] = False
                    log_configuration_override_details[
                        self.CONST_ENABLE_AUTO_LOG_CREATION
                    ] = False

        if step_override_details:
            step_override_details_list = []
            for step in step_override_details:
                step_detail = {}
                step_detail["stepName"] = step["step_name"]
                step_detail["stepConfigurationDetails"] = self._standardize_spec(
                    step["step_configuration_details"]
                )
                step_override_details_list.append(step_detail)
            pipeline_details[
                self.CONST_STEP_OVERRIDE_DETAILS
            ] = step_override_details_list

        if free_form_tags:
            pipeline_details[self.CONST_FREEFROM_TAGS] = free_form_tags

        if defined_tags:
            pipeline_details[self.CONST_DEFINED_TAGS] = defined_tags

        if system_tags:
            pipeline_details[self.CONST_SYSTEM_TAGS] = system_tags

    # TODO: Needs to improve the validation logic
    # Ticket: https://jira.oci.oraclecorp.com/browse/ODSC-31996
    # @classmethod
    # def from_yaml(cls, uri: str) -> "Pipeline":
    #     pipeline_schema = {}
    #     schema_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "schema")
    #     with open(
    #         os.path.join(schema_path, "pipeline_schema.json")
    #     ) as pipeline_schema_file:
    #         pipeline_schema = json.load(pipeline_schema_file)

    #     cs_step_schema = {}
    #     with open(
    #         os.path.join(schema_path, "cs_step_schema.json")
    #     ) as cs_step_schema_file:
    #         cs_step_schema = json.load(cs_step_schema_file)

    #     ml_step_schema = {}
    #     with open(
    #         os.path.join(schema_path, "ml_step_schema.json")
    #     ) as ml_step_schema_file:
    #         ml_step_schema = json.load(ml_step_schema_file)

    #     yaml_dict = yaml.load(Pipeline._read_from_file(uri=uri))

    #     pipeline_validator = Validator(pipeline_schema)
    #     if not pipeline_validator.validate(yaml_dict):
    #         raise ValueError(pipeline_validator.errors)

    #     step_details = yaml_dict["spec"]["stepDetails"]
    #     if len(step_details) == 0:
    #         raise ValueError("Pipeline must have at least one step.")

    #     ml_step_validator = Validator(ml_step_schema)
    #     cs_step_validator = Validator(cs_step_schema)
    #     for step in step_details:
    #         if not ml_step_validator.validate(step) and not cs_step_validator.validate(
    #             step
    #         ):
    #             if ml_step_validator.errors:
    #                 raise ValueError(ml_step_validator.errors)
    #             else:
    #                 raise ValueError(cs_step_validator.errors)

    #     return super().from_yaml(uri=uri)

    @classmethod
    def list(cls, compartment_id: Optional[str] = None, **kwargs) -> List["Pipeline"]:
        """
        List pipelines in a given compartment.

        Parameters
        ----------
        compartment_id: (str, optional). Defaults to None.
            The OCID of compartment.
            If `None`, the value will be taken from the environment variables.
        kwargs
            Additional keyword arguments for filtering pipelines.
            - project_id: str
            - lifecycle_state: str. Allowed values: "CREATING", "ACTIVE", "DELETING", "FAILED", "DELETED"
            - created_by: str
            - limit: int

        Returns
        -------
        List[Pipeline]
            The list of pipelines.
        """
        result = []
        for item in DataSciencePipeline.list_resource(compartment_id, **kwargs):
            pipeline = item.build_ads_pipeline()
            pipeline._populate_step_artifact_content()
            result.append(pipeline)
        return result

    def run_list(self, **kwargs) -> List[PipelineRun]:
        """Gets a list of runs of the pipeline.

        Returns
        -------
        List[PipelineRun]
            A list of pipeline run instances.
        """
        return PipelineRun.list(
            compartment_id=self.compartment_id, pipeline_id=self.id, **kwargs
        )

    @property
    def status(self) -> Optional[str]:
        """Status of the pipeline.

        Returns
        -------
        str
            Status of the pipeline.
        """
        if self.data_science_pipeline:
            return self.data_science_pipeline.lifecycle_state
        return None


class DataSciencePipeline(OCIDataScienceMixin, oci.data_science.models.Pipeline):
    @classmethod
    def from_ocid(cls, ocid: str) -> "DataSciencePipeline":
        """Gets a datascience pipeline by OCID.

        Parameters
        ----------
        ocid: str
            The OCID of the datascience pipeline.

        Returns
        -------
        DataSciencePipeline
            An instance of DataSciencePipeline.
        """
        return super().from_ocid(ocid)

    def build_ads_pipeline(self) -> "Pipeline":
        """Builds an ADS pipeline from OCI datascience pipeline.

        Returns
        -------
        Pipeline:
            ADS Pipeline instance.
        """
        pipeline_details = self.to_dict()
        ads_pipeline = Pipeline(pipeline_details["displayName"])
        ads_pipeline.data_science_pipeline = self

        for key in pipeline_details:
            if key in ads_pipeline.attribute_set:
                if key == "stepDetails":
                    step_details = []
                    for step in pipeline_details[key]:
                        step_details.append(self.build_ads_pipeline_step(step))
                    ads_pipeline.set_spec(key, step_details)
                elif key in ["freeformTags", "systemTags", "definedTags"]:
                    ads_pipeline.set_spec(key, pipeline_details[key])
                elif type(pipeline_details[key]) is dict:
                    for attribute in pipeline_details[key]:
                        ads_pipeline.set_spec(
                            attribute, pipeline_details[key][attribute]
                        )
                else:
                    ads_pipeline.set_spec(key, pipeline_details[key])

        dag_list = ads_pipeline._convert_step_details_to_dag(
            pipeline_details["stepDetails"]
        )
        ads_pipeline.set_spec(Pipeline.CONST_DAG, dag_list)

        return ads_pipeline

    def build_ads_pipeline_step(self, step: Dict) -> "PipelineStep":
        """Builds an ADS pipeline step from OCI pipeline response.

        Parameters
        ----------
        step: dict
            A dictionary that contains the information of a pipeline step.

        Returns
        -------
        Pipeline:
            ADS PipelineStep instance.
        """
        ads_pipeline_step = PipelineStep(step["stepName"])

        for key in step:
            if key in ads_pipeline_step.attribute_set:
                infrastructure = DataScienceJob()
                if key == ads_pipeline_step.CONST_STEP_INFRA_CONFIG_DETAILS:
                    for attribute in step[key]:
                        infrastructure.set_spec(attribute, step[key][attribute])
                    ads_pipeline_step.set_spec(
                        ads_pipeline_step.CONST_INFRASTRUCTURE, infrastructure
                    )
                elif key == ads_pipeline_step.CONST_STEP_CONFIG_DETAILS:
                    if step["stepType"] == "CUSTOM_SCRIPT":
                        job_configuration_details_dict = {}
                        for attribute in step[key]:
                            job_configuration_details_dict[
                                infrastructure.CONST_JOB_TYPE
                            ] = "DEFAULT"
                            job_configuration_details_dict[attribute] = step[key][
                                attribute
                            ]
                        dsc_job = DSCJob(
                            job_configuration_details=job_configuration_details_dict
                        )
                        runtime = DataScienceJobRuntimeManager(infrastructure).extract(
                            dsc_job
                        )
                        ads_pipeline_step.set_spec(
                            ads_pipeline_step.CONST_RUNTIME, runtime
                        )
                    else:
                        for attribute in step[key]:
                            ads_pipeline_step.set_spec(attribute, step[key][attribute])
                else:
                    ads_pipeline_step.set_spec(key, step[key])
        return ads_pipeline_step

    def create(self, step_details: List, delete_if_fail: bool) -> str:
        """Creates an OCI pipeline.

        Parameters
        ----------
        step_details: list
            List of pipeline step details.

        Returns
        -------
        str:
            The id of OCI pipeline.
        """
        response = self.client.create_pipeline(
            self.to_oci_model(oci.data_science.models.CreatePipelineDetails)
        )
        self.update_from_oci_model(response.data)
        try:
            self.upload_artifact(step_details)
        except Exception as ex:
            if delete_if_fail:
                self.delete(self.id)
            raise ex
        self.step_details = step_details
        return self

    def upload_artifact(self, step_details: List) -> "DataSciencePipeline":
        """Uploads artifacts to pipeline.

        Parameters
        ----------
        step_details: list
            List of pipeline step details.

        Returns
        -------
        DataSciencePipeline:
            DataSciencePipeline instance.
        """
        for step in step_details:
            if step.runtime:
                payload = DataScienceJobRuntimeManager(step.infrastructure).translate(
                    step.runtime
                )
                target_artifact = payload["artifact"]
                if issubclass(target_artifact.__class__, Artifact):
                    with target_artifact as artifact:
                        self.create_step_artifact(artifact.path, step.name)
                else:
                    self.create_step_artifact(target_artifact, step.name)
        return self

    def create_step_artifact(
        self, artifact_path: str, step_name: str
    ) -> "DataSciencePipeline":
        """Creates step artifact.

        Parameters
        ----------
        artifact_path: str
            Local path to artifact.
        step_name: str
            Pipeline step name.

        Returns
        -------
        DataSciencePipeline:
            DataSciencePipeline instance.
        """
        with fsspec.open(artifact_path, "rb") as f:
            self.client.create_step_artifact(
                self.id,
                step_name,
                f,
                content_disposition=f"attachment; filename={os.path.basename(artifact_path)}",
            )
        return self

    def run(
        self, pipeline_details: Dict, service_logging: OCILog = None
    ) -> "PipelineRun":
        """Runs an OCI pipeline.

        Parameters
        ----------
        pipeline_details: dict
            A dictionary that contains pipeline details.
        service_logging: OCILog instance.
            The OCILog instance.

        Returns
        -------
        PipelineRun:
            PipelineRun instance.
        """
        data_science_pipeline_run = PipelineRun(**pipeline_details)
        if service_logging:
            data_science_pipeline_run._set_service_logging_resource(service_logging)
        data_science_pipeline_run.create()

        return data_science_pipeline_run

    def delete(
        self,
        id: str,
        operation_kwargs: Dict = DEFAULT_OPERATION_KWARGS,
        waiter_kwargs: Dict = DEFAULT_WAITER_KWARGS,
    ) -> "DataSciencePipeline":
        """Deletes an OCI pipeline.

        Parameters
        ----------
        id: str
            The ocid of pipeline.
        Parameters
        ----------
        operation_kwargs: dict, optional
            The operational kwargs to be executed when deleting the pipeline.
            Defaults to: {"delete_related_pipeline_runs": True, "delete_related_job_runs": True},
            which will delete the corresponding pipeline runs and job runs.

            The allowed keys are:
            * "delete_related_pipeline_runs": bool, to specify whether to delete related
            PipelineRuns or not.
            * "delete_related_job_runs": bool, to specify whether to delete related JobRuns or not.
            * "allow_control_chars": bool, to indicate whether or not this request should
            allow control characters in the response object. By default, the response will not
            allow control characters in strings
            * "retry_strategy": obj, to apply to this specific operation/call. This will
            override any retry strategy set at the client-level. This should be one of the
            strategies available in the :py:mod:`~oci.retry` module. This operation will not retry
            by default, users can also use the convenient :py:data:`~oci.retry.DEFAULT_RETRY_STRATEGY`
            provided by the SDK to enable retries for it. The specifics of the default retry strategy
            are described `here <https://docs.oracle.com/en-us/iaas/tools/python/latest/sdk_behaviors/retries.html>`__.
            To have this operation explicitly not perform any retries, pass an instance of :py:class:`~oci.retry.NoneRetryStrategy`.
            * "if_match": str, for optimistic concurrency control. In the PUT or DELETE call
            for a resource, set the `if-match` parameter to the value of the etag from a previous
            GET or POST response for that resource. The resource is updated or deleted only if the
            `etag` you provide matches the resource's current `etag` value.
            * "opc_request_id": str, unique Oracle assigned identifier for the request. If you need
            to contact Oracle about a particular request, then provide the request ID.

        waiter_kwargs: dict, optional
            The waiter kwargs to be passed when deleting the pipeline.
            Defaults to: {"max_wait_seconds": 1800}, which will allow a maximum wait time to 1800 seconds to delete the pipeline.
            The allowed keys are:
            * "max_wait_seconds": int, the maximum time to wait, in seconds.
            * "max_interval_seconds": int, the maximum interval between queries, in seconds.
            * "succeed_on_not_found": bool, to determine whether or not the waiter should return
            successfully if the data we're waiting on is not found (e.g. a 404 is returned from the service).
            This defaults to False and so a 404 would cause an exception to be thrown by this function.
            Setting it to True may be useful in scenarios when waiting for a resource to be
            terminated/deleted since it is possible that the resource would not be returned by the a GET call anymore.
            * "wait_callback": A function which will be called each time that we have to do an initial
            wait (i.e. because the property of the resource was not in the correct state,
            or the ``evaluate_response`` function returned False). This function should take two
            arguments - the first argument is the number of times we have checked the resource,
            and the second argument is the result of the most recent check.
            * "fetch_func": A function to be called to fetch the updated state from the server.
            This can be used if the call to check for state needs to be more complex than a single
            GET request. For example, if the goal is to wait until an item appears in a list,
            fetch_func can be a function that paginates through a full list on the server.

        Returns
        -------
        DataSciencePipeline:
            DataSciencePipeline instance.
        """
        self.client_composite.delete_pipeline_and_wait_for_state(
            pipeline_id=id,
            wait_for_states=[
                oci.data_science.models.WorkRequest.STATUS_SUCCEEDED,
                oci.data_science.models.WorkRequest.STATUS_FAILED,
            ],
            operation_kwargs=operation_kwargs,
            waiter_kwargs=waiter_kwargs,
        )
        return self.sync()
